import logging
import math
from typing import Sequence

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import dim_compress, libentry, libtuner
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(
    f'flag_gems.runtime.backend._mthreads.ops.{__name__.split(".")[-1]}'
)

NAIVE_REDUCTION_CONFIGS = [
    triton.Config({"BLOCK_M": 4, "BLOCK_N": 1024}, num_warps=4),
    triton.Config({"BLOCK_M": 8, "BLOCK_N": 1024}, num_warps=4),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 1024}, num_warps=8),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 512}, num_warps=8),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 256}, num_warps=4),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 128}, num_warps=4),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 64}, num_warps=4),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 256}, num_warps=4),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 128}, num_warps=4),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 256}, num_warps=8),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 128}, num_warps=4),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 64}, num_warps=4),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 64}, num_warps=8),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 32}, num_warps=4),
    triton.Config({"BLOCK_M": 128, "BLOCK_N": 256}, num_warps=8),
    triton.Config({"BLOCK_M": 256, "BLOCK_N": 256}, num_warps=8),
]


@triton.jit
def reduce_all(a, b):
    return a and b


@triton.autotune(configs=NAIVE_REDUCTION_CONFIGS, key=["M", "N"])
@triton.jit
def all_kernel_dim_strided(
    inp,
    out,
    M,
    N,
    INNER,
    STRIDE_OUTER,
    STRIDE_REDUCE,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    pid = tl.program_id(0)
    rows = pid * BLOCK_M + tl.arange(0, BLOCK_M)
    rows = rows.to(tl.int64)
    row_mask = rows < M

    outer_idx = rows // INNER
    inner_idx = rows % INNER
    base_ptr = inp + outer_idx * STRIDE_OUTER + inner_idx

    acc = tl.full([BLOCK_M, BLOCK_N], value=1, dtype=tl.int1)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)
        cols = cols.to(tl.int64)
        col_mask = cols < N
        mask = row_mask[:, None] and col_mask[None, :]
        vals = tl.load(
            base_ptr[:, None] + cols[None, :] * STRIDE_REDUCE, mask, other=1.0
        )
        acc = acc and (vals != 0)
    all_val = tl.reduce(acc, axis=1, combine_fn=reduce_all)
    tl.store(out + rows, all_val, mask=row_mask)


def _flatten_dim(shape: Sequence[int], dim: int):
    dim = dim % len(shape)
    n = shape[dim]
    inner = math.prod(shape[dim + 1 :]) if dim + 1 < len(shape) else 1
    outer = math.prod(shape[:dim]) if dim > 0 else 1
    return dim, n, inner, outer


def triton_all_dim_strided(
    inp: torch.Tensor, dim: int, keepdim: bool = False
) -> torch.Tensor:
    dim = dim % inp.ndim
    shape = list(inp.shape)
    dim, n, inner, outer = _flatten_dim(shape, dim)
    m = outer * inner

    stride = inp.stride()
    stride_reduce = stride[dim]
    stride_outer = stride_reduce * n

    out_flat = torch.empty((m,), dtype=torch.bool, device=inp.device)
    grid = lambda meta: (triton.cdiv(m, meta["BLOCK_M"]),)
    all_kernel_dim_strided[grid](
        inp,
        out_flat,
        m,
        n,
        inner,
        stride_outer,
        stride_reduce,
    )

    shape[dim] = 1
    out = out_flat.view(shape)
    if not keepdim:
        out = out.squeeze(dim=dim)
    return out


@libentry()
@libtuner(
    configs=runtime.get_tuned_config("naive_reduction"),
    key=["M", "N"],
)
@triton.jit
def all_kernel_dim(
    inp,
    out,
    M,
    N,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    pid = tle.program_id(0)
    rows = pid * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    inp = inp + rows * N
    out = out + rows
    row_mask = rows < M

    _all = tl.full([BLOCK_M, BLOCK_N], value=1, dtype=tl.int1)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(inp + cols, mask, other=1.0)
        _all = _all and (a != 0)
    all = tl.reduce(_all, axis=1, combine_fn=reduce_all)
    tl.store(out, all[:, None], row_mask)


@libentry()
@triton.jit
def all_kernel_1(
    inp,
    mid,
    n_elements,
    mid_size,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < n_elements
    inp_val = tl.load(inp_ptrs, mask=mask, other=1.0)
    all_val = tl.reduce(inp_val != 0, axis=0, combine_fn=reduce_all)
    mid_ptr = mid + pid
    tl.store(mid_ptr, all_val)


@libentry()
@triton.jit
def all_kernel_2(mid, out, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid + offset
    mask = offset < MID_SIZE
    mid_val = tl.load(mid_ptrs, mask=mask, other=1).to(tl.int1)
    all_val = tl.reduce(mid_val, axis=0, combine_fn=reduce_all)
    tl.store(out, all_val)


def all(inp):
    logger.debug("GEMS_MTHREADS ALL")
    n_elements = inp.numel()
    block_size = triton.next_power_of_2(math.ceil(math.sqrt(n_elements)))
    block_size = min(block_size * 2, 4096, triton.next_power_of_2(n_elements))
    mid_size = triton.cdiv(n_elements, block_size)
    block_mid = triton.next_power_of_2(mid_size)

    mid = torch.empty((mid_size,), dtype=torch.bool, device=inp.device)
    out = torch.empty([], dtype=torch.bool, device=inp.device)

    with torch_device_fn.device(inp.device):
        all_kernel_1[(mid_size, 1)](inp, mid, n_elements, mid_size, block_size)
        all_kernel_2[(1, 1)](mid, out, mid_size, block_mid)

    return out


def all_dim(inp, dim=None, keepdim=False):
    logger.debug("GEMS_MTHREADS ALL DIM")
    if dim is None:
        out = all(inp)
        if keepdim:
            out = torch.reshape(out, [1] * inp.ndim)
        return out

    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    dim = dim % inp.ndim

    with torch_device_fn.device(inp.device):
        return triton_all_dim_strided(inp, dim=dim, keepdim=keepdim)


def all_dims(inp, dim=None, keepdim=False):
    logger.debug("GEMS_MTHREADS ALL DIMS")

    if dim is None or isinstance(dim, int):
        return all_dim(inp, dim=dim, keepdim=keepdim)
    assert ((i >= -inp.ndim and i < inp.ndim) for i in dim), "Invalid dim"

    shape = list(inp.shape)
    dim = [d % inp.ndim for d in dim]
    inp = dim_compress(inp, dim)
    N = 1
    for i in dim:
        N *= shape[i]
        shape[i] = 1
    M = inp.numel() // N

    out = torch.empty(shape, dtype=torch.bool, device=inp.device)

    grid = lambda meta: (triton.cdiv(M, meta["BLOCK_M"]),)
    with torch_device_fn.device(inp.device):
        all_kernel_dim[grid](inp, out, M, N)
    if not keepdim:
        out = out.squeeze(dim=dim)
    return out


__all__ = ["all", "all_dim", "all_dims"]



# This custom op requires musa device capability >= 31.
# We determine whether to enable this op by distinguish the op registration for different arch.

import logging

import torch
import triton
import triton.language as tl

from flag_gems.utils import pointwise_dynamic, tl_extra_shim

logger = logging.getLogger(
    f'flag_gems.runtime.backend._mthreads.ops.{__name__.split(".")[-1]}'
)
erf = tl_extra_shim.erf
exp = tl_extra_shim.exp
pow = tl_extra_shim.pow
fast_tanh = tl_extra_shim.fast_tanh
fast_gelu = tl_extra_shim.fast_gelu


@pointwise_dynamic(promotion_methods=[(0, "DEFAULT")])
@triton.jit
def gelu_none(x):
    return fast_gelu(x.to(tl.float32))


@pointwise_dynamic(promotion_methods=[(0, "DEFAULT")])
@triton.jit
def gelu_tanh(x):
    output = (
        0.5
        * x
        * (1 + fast_tanh(x * 0.79788456 * (1 + 0.044715 * pow(x.to(tl.float32), 2))))
    )
    return output


@pointwise_dynamic(promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def gelu_backward_none(x, dy):
    scale1: tl.constexpr = 0.7071067811  # 1 / math.sqrt(2)
    scale2: tl.constexpr = 0.3989422803  # 1 / math.sqrt(2 * math.pi)
    x_fp32 = x.to(tl.float32)
    dydx = (
        scale2 * x_fp32 * exp(-pow(scale1 * x_fp32, 2))
        + 0.5 * erf(scale1 * x_fp32)
        + 0.5
    )
    dx = dydx * dy
    return dx


@pointwise_dynamic(promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def gelu_backward_tanh(x, dy):
    x_fp32 = x.to(tl.float32)
    # 0.79788456 = math.sqrt(2 / math.pi)
    tanh_out = fast_tanh(0.79788456 * x * (1 + 0.044715 * pow(x_fp32, 2)))
    dydx = 0.5 * x * (
        (1 - pow(tanh_out, 2)) * (0.79788456 + 0.1070322243 * pow(x_fp32, 2))
    ) + 0.5 * (1 + tanh_out)
    dx = dydx * dy
    return dx


class Gelu(torch.autograd.Function):
    @staticmethod
    def forward(ctx, A, approximate):
        logger.debug("GEMS_MTHREADS GELU FORWARD")
        if approximate == "tanh":
            out = gelu_tanh(A)
        else:
            out = gelu_none(A)
        ctx.save_for_backward(A)
        ctx.approximate = approximate
        return out

    @staticmethod
    def backward(ctx, out_grad):
        logger.debug("GEMS_MTHREADS GELU BACKWARD")
        (inp,) = ctx.saved_tensors
        approximate = ctx.approximate
        if approximate == "tanh":
            in_grad = gelu_backward_tanh(inp, out_grad)
        else:
            in_grad = gelu_backward_none(inp, out_grad)
        return in_grad, None


def gelu(A, *, approximate="none"):
    return Gelu.apply(A, approximate)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project
import torch

from vllm.distributed.parallel_state import GroupCoordinator
from vllm.triton_utils import tl, triton


@triton.jit
def _correct_attn_cp_out_kernel(
    outputs_ptr,
    new_output_ptr,
    lses_ptr,
    vlse_ptr,
    outputs_stride_B,
    outputs_stride_H,
    outputs_stride_D,
    lses_stride_N,
    lses_stride_B,
    lses_stride_H,
    lse_idx,
    HEAD_DIM: tl.constexpr,
    N_ROUNDED: tl.constexpr,
    IS_BASE_E: tl.constexpr,
):
    """
    Apply the all-gathered lses to correct each local rank's attention
    output. we still need perform a cross-rank reduction to obtain the
    final attention output.

    Args:
        outputs_ptr (triton.PointerType):
            Pointer to input tensor of shape [ B, H, D ]
        lses_ptr (triton.PointerType):
            Pointer to input tensor of shape [ N, B, H ]
        new_output_ptr (triton.PointerType):
            Pointer to output tensor of shape [ B, H, D ]
        vlse_ptr (triton.PointerType):
            Pointer to output tensor of shape [ B, H ]
    """
    batch_idx = tl.program_id(axis=0).to(tl.int64)
    head_idx = tl.program_id(axis=1).to(tl.int64)
    d_offsets = tl.arange(0, HEAD_DIM)
    num_n_offsets = tl.arange(0, N_ROUNDED)

    # shape = [N]
    lse_offsets = (
        num_n_offsets * lses_stride_N
        + batch_idx * lses_stride_B
        + head_idx * lses_stride_H
    )

    # calc final lse
    lse = tl.load(lses_ptr + lse_offsets)
    lse = tl.where((lse != lse) | (lse == float("inf")), -float("inf"), lse)
    lse_max = tl.max(lse, axis=0)
    lse_max = tl.where(lse_max == -float("inf"), 0, lse_max)
    lse -= lse_max
    if IS_BASE_E:
        lse_exp = tl.exp(lse)
        lse_acc = tl.sum(lse_exp, axis=0)
        lse = tl.log(lse_acc)
    else:
        lse_exp = tl.exp2(lse)
        lse_acc = tl.sum(lse_exp, axis=0)
        lse = tl.log2(lse_acc)
    lse += lse_max

    lse_offsets = batch_idx * lses_stride_B + head_idx * lses_stride_H
    tl.store(vlse_ptr + lse_offsets, lse)

    # shape = [D]
    output_offsets = (
        batch_idx * outputs_stride_B
        + head_idx * outputs_stride_H
        + d_offsets * outputs_stride_D
    )

    # correct output
    lse_offset = (
        lse_idx * lses_stride_N + batch_idx * lses_stride_B + head_idx * lses_stride_H
    )
    lse_tmp = tl.load(lses_ptr + lse_offset)
    lse_finally = lse_tmp - lse
    lse_finally = tl.where(
        (lse_finally != lse_finally) | (lse_finally == float("inf")),
        -float("inf"),
        lse_finally,
    )
    factor = tl.exp(lse_finally) if IS_BASE_E else tl.exp2(lse_finally)
    output = tl.load(outputs_ptr + output_offsets)
    output = output * factor

    tl.store(new_output_ptr + output_offsets, output)


class CPTritonContext:
    """The CPTritonContext is used to avoid recompilation of the Triton JIT."""

    def __init__(self):
        self.inner_kernel = None

    def call_kernel(self, kernel, grid, *regular_args, **const_args):
        if self.inner_kernel is None:
            self.inner_kernel = kernel[grid](*regular_args, **const_args)
        else:
            self.inner_kernel[grid](*regular_args)


def correct_attn_out(
    out: torch.Tensor,
    lses: torch.Tensor,
    cp_rank: int,
    ctx: CPTritonContext,
    is_lse_base_on_e: bool = True,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Correct the attention output using the all-gathered lses.

    Args:
        out: Tensor of shape [ B, H, D ]
        lses: Tensor of shape [ N, B, H ]
        cp_rank: Current rank in the context-parallel group
        ctx: Triton context to avoid recompilation

    Returns:
        Tuple of (out, lse) with corrected attention and final log-sum-exp.
    """
    if ctx is None:
        ctx = CPTritonContext()

    # --- Normalize to 3D views ---
    if out.ndim == 4 and out.shape[1] == 1:
        out = out.squeeze(1)
    assert out.ndim == 3, f"expected out [B,H,D] or [B,1,H,D], got {tuple(out.shape)}"

    if lses.ndim == 4 and lses.shape[-1] == 1:
        lses = lses.squeeze(-1)
    if lses.ndim == 4 and lses.shape[1] == 1:
        lses = lses.squeeze(1)
    assert lses.ndim == 3, (
        f"expected lses [N,B,H] (optionally with a 1-sized extra dim), "
        f"got {tuple(lses.shape)}"
    )

    B, H, D = out.shape
    N = lses.shape[0]

    # Strides after we normalized shapes to 3-D views.  The kernel computes
    # offsets for `vlse_ptr` using lses_stride_B/H, so the output buffer must
    # have the same B/H stride layout as a slice of `lses`.
    o_sB, o_sH, o_sD = out.stride()
    l_sN, l_sB, l_sH = lses.stride()

    # Allocate LSE with the same B/H strides as `lses` so writes land correctly
    # even when `lses` is a non-contiguous view (e.g., 4-D to 3-D squeeze).
    lse = torch.empty_strided(
        (B, H), (l_sB, l_sH), device=lses.device, dtype=lses.dtype
    )

    # Kernel launch config
    grid = (B, H, 1)

    regular_args = (
        out,
        out,
        lses,
        lse,
        o_sB,
        o_sH,
        o_sD,
        l_sN,
        l_sB,
        l_sH,
        cp_rank,
    )
    const_args = {"HEAD_DIM": D, "N_ROUNDED": N, "IS_BASE_E": is_lse_base_on_e}
    ctx.call_kernel(_correct_attn_cp_out_kernel, grid, *regular_args, **const_args)
    return out, lse


def _cp_lse_common(
    cp_attn_out: torch.Tensor,
    cp_attn_lse: torch.Tensor,
    cp_group: GroupCoordinator,
    ctx: CPTritonContext | None = None,
    is_lse_base_on_e=True,
):
    """
    cp_attn_out: [ B, H, D ]
    cp_attn_lse: [ B, H ]
    """
    if cp_group.world_size == 1:
        return cp_attn_out

    if ctx is None:
        ctx = CPTritonContext()

    lses = torch.empty(
        (cp_group.world_size,) + cp_attn_lse.shape,
        dtype=cp_attn_lse.dtype,
        device=cp_attn_lse.device,
    )

    cp_attn_lse = cp_attn_lse.contiguous()
    lses = cp_group.all_gather(cp_attn_lse, dim=0).view_as(lses)
    out, lse = correct_attn_out(
        cp_attn_out,
        lses,
        cp_group.rank_in_group,
        ctx,
        is_lse_base_on_e=is_lse_base_on_e,
    )
    return out, lse


def cp_lse_ag_out_rs(
    cp_attn_out: torch.Tensor,
    cp_attn_lse: torch.Tensor,
    cp_group: GroupCoordinator,
    ctx: CPTritonContext | None = None,
    return_lse: bool = False,
    is_lse_base_on_e=True,
):
    """
    cp_attn_out: [ B, H, D ]
    cp_attn_lse: [ B, H ]
    """
    out, lse = _cp_lse_common(
        cp_attn_out, cp_attn_lse, cp_group, ctx=ctx, is_lse_base_on_e=is_lse_base_on_e
    )
    out = cp_group.reduce_scatter(out, dim=1)

    if return_lse:
        cp_num_heads = lse.shape[1] // cp_group.world_size
        cp_rank = cp_group.rank_in_group
        lse = lse[:, cp_num_heads * cp_rank : cp_num_heads * (cp_rank + 1)]
        return out, lse
    return out


def cp_lse_ag_out_ar(
    cp_attn_out: torch.Tensor,
    cp_attn_lse: torch.Tensor,
    cp_group: GroupCoordinator,
    ctx: CPTritonContext | None = None,
    return_lse: bool = False,
    is_lse_base_on_e=True,
):
    """
    cp_attn_out: [ B, H, D ]
    cp_attn_lse: [ B, H ]
    """
    out, lse = _cp_lse_common(
        cp_attn_out, cp_attn_lse, cp_group, ctx=ctx, is_lse_base_on_e=is_lse_base_on_e
    )
    out = cp_group.all_reduce(out)

    if return_lse:
        return out, lse
    return out


@triton.jit
def _pack_seq_kernel(
    x_ptr,  # [N, D]
    out_ptr,  # [B, Lmax, D]
    lengths_ptr,  # *i32, [B]
    N: tl.constexpr,
    D: tl.constexpr,
    Lmax: tl.constexpr,
    PAD_VALUE: tl.constexpr,
    BLOCK_T: tl.constexpr,  # timesteps per program
    BLOCK_D: tl.constexpr,  # features per program
):
    pid_b = tl.program_id(0)  # batch id
    pid_t = tl.program_id(1)  # block over time dimension
    pid_d = tl.program_id(2)  # block over feature dimension
    off_t = pid_t * BLOCK_T + tl.arange(0, BLOCK_T)  # [BLOCK_T]
    off_d = pid_d * BLOCK_D + tl.arange(0, BLOCK_D)  # [BLOCK_D]

    # Compute start index and sequence length from cumulative lengths
    in_start = 0
    for i in range(pid_b):
        in_start += tl.load(lengths_ptr + i)
    seq_len = tl.load(lengths_ptr + pid_b)

    # valid time positions for this block
    t_mask = off_t < Lmax

    # compute input row indices for valid (b, t)
    in_row = in_start + off_t
    valid_row = (off_t < seq_len) & t_mask

    # Pointers
    # x_ptr: row-major [N, D]
    x_row_ptr = x_ptr + in_row[:, None] * D + off_d[None, :]

    # out_ptr: row-major [B, Lmax, D]
    out_row_ptr = out_ptr + (pid_b * Lmax + off_t)[:, None] * D + off_d[None, :]

    # Initialize with PAD (cast will occur as needed based on out_ptr dtype)
    d_mask = off_d[None, :] < D
    pad_vals = tl.full([BLOCK_T, BLOCK_D], PAD_VALUE, tl.float32)
    tl.store(out_row_ptr, pad_vals, mask=t_mask[:, None] & d_mask)

    # Load & write only where within seq_len
    x_vals = tl.load(x_row_ptr, mask=valid_row[:, None] & d_mask)
    tl.store(out_row_ptr, x_vals, mask=valid_row[:, None] & d_mask)


def pack_seq_triton(
    x: torch.Tensor,
    lengths: torch.Tensor,
    pad_value: float = -float("inf"),
    block_t: int = 64,
    block_d: int = 64,
) -> torch.Tensor:
    """
    Pack sequences of different lengths into a batched tensor.

    Args:
        x: [N, ...] - input tensor where N is total number of tokens
        lengths: [B] - sequence lengths for each batch
        pad_value: value to use for padding
        block_t: block size for time dimension
        block_d: block size for feature dimension

    Returns:
        packed: [B, Lmax, ...] - packed tensor
    """

    # Handle multi-dimensional input by reshaping to (N, -1)
    original_shape = x.shape
    if len(original_shape) > 2:
        N = original_shape[0]
        x_reshaped = x.reshape(N, -1)
        D = x_reshaped.shape[1]
    else:
        N, D = x.shape
        x_reshaped = x

    B = lengths.numel()
    Lmax = int(lengths.max().item())

    # Starts are computed inside the kernel from lengths

    out = torch.empty((B, Lmax, D), device=x.device, dtype=x.dtype)

    grid = (B, triton.cdiv(Lmax, block_t), triton.cdiv(D, block_d))
    _pack_seq_kernel[grid](
        x_reshaped,
        out,
        lengths.int(),
        N,
        D,
        Lmax,
        PAD_VALUE=float(pad_value),
        BLOCK_T=block_t,
        BLOCK_D=block_d,
        num_warps=4,
        num_stages=2,
    )

    # Reshape output back to original dimensions (except first dimension)
    if len(original_shape) > 2:
        output_shape = (B, Lmax) + original_shape[1:]
        out = out.reshape(output_shape)

    return out


@triton.jit
def _unpack_seq_triton_kernel(
    packed_ptr,  # [B, Lmax, D]
    out_ptr,  # [N, D]
    lengths_ptr,  # *i32, [B]
    B: tl.constexpr,
    Lmax: tl.constexpr,
    D: tl.constexpr,
    BLOCK_T: tl.constexpr,  # timesteps per program
    BLOCK_D: tl.constexpr,  # features per program
):
    pid_b = tl.program_id(0)  # batch id
    pid_t = tl.program_id(1)  # block over time dimension
    pid_d = tl.program_id(2)  # block over feature dimension
    off_t = pid_t * BLOCK_T + tl.arange(0, BLOCK_T)  # [BLOCK_T]
    off_d = pid_d * BLOCK_D + tl.arange(0, BLOCK_D)  # [BLOCK_D]

    # bounds: compute start from cumulative lengths
    in_start = 0
    for i in range(pid_b):
        in_start += tl.load(lengths_ptr + i)
    seq_len = tl.load(lengths_ptr + pid_b)

    # valid time positions for this block
    t_mask = off_t < Lmax
    valid_row = (off_t < seq_len) & t_mask

    # compute output row indices for valid (b, t)
    out_row = in_start + off_t

    # Pointers
    # packed_ptr: row-major [B, Lmax, D]
    packed_row_ptr = packed_ptr + (pid_b * Lmax + off_t)[:, None] * D + off_d[None, :]

    # out_ptr: row-major [N, D]
    out_row_ptr = out_ptr + out_row[:, None] * D + off_d[None, :]

    # Load from packed tensor and store to output
    d_mask = off_d[None, :] < D
    packed_vals = tl.load(packed_row_ptr, mask=valid_row[:, None] & d_mask)
    tl.store(out_row_ptr, packed_vals, mask=valid_row[:, None] & d_mask)


def unpack_seq_triton(
    packed_tensor: torch.Tensor,
    lengths: torch.Tensor,
    block_t: int = 64,
    block_d: int = 64,
) -> torch.Tensor:
    """
    Unpack a packed decode query tensor back to the original format.
    Efficient Triton implementation.

    Args:
        packed_tensor: [B, Lmax, ...] - packed tensor from pack_seq_triton
        lengths: [B] - sequence lengths for each batch
        block_t: block size for time dimension
        block_d: block size for feature dimension

    Returns:
        unpacked_tensor: [N, ...] where N = sum(lengths)
    """

    # Handle multi-dimensional input by reshaping to (B, Lmax, -1)
    original_shape = packed_tensor.shape
    if len(original_shape) > 3:
        B, Lmax = original_shape[:2]
        packed_reshaped = packed_tensor.reshape(B, Lmax, -1)
        D = packed_reshaped.shape[2]
    else:
        B, Lmax, D = packed_tensor.shape
        packed_reshaped = packed_tensor

    # Calculate total number of elements
    N = int(lengths.sum().item())

    out = torch.empty((N, D), device=packed_tensor.device, dtype=packed_tensor.dtype)

    grid = (B, triton.cdiv(Lmax, block_t), triton.cdiv(D, block_d))
    _unpack_seq_triton_kernel[grid](
        packed_reshaped,
        out,
        lengths.int(),
        B,
        Lmax,
        D,
        BLOCK_T=block_t,
        BLOCK_D=block_d,
        num_warps=4,
        num_stages=2,
    )

    # Reshape output back to original dimensions (except first dimension)
    if len(original_shape) > 3:
        output_shape = (N,) + original_shape[2:]
        out = out.reshape(output_shape)

    return out



# LayerNorm - Pure Triton
# Source: FlagGems (Apache 2.0)
# NeuroBrix - NVIDIA Common

import torch
import triton
import triton.language as tl

@triton.jit
def prev_multiple_of(a, b):
    # the largest x<a that x%b ==0
    return tl.cdiv(a, b) * b - b

@triton.jit
def layer_norm_persistent_kernel(
    in_ptr, out_ptr, weight_ptr, bias_ptr,
    out_mean_ptr, out_rstd_ptr,
    M, N, eps,
    TILE_N: tl.constexpr,
):
    pid = tl.program_id(0)
    n_offsets = tl.arange(0, TILE_N)
    mask = n_offsets < N

    x = tl.load(in_ptr + pid * N + n_offsets, mask, other=0.0).to(tl.float32)
    m = tl.sum(x, axis=0) / N
    d = x - m
    s = tl.where(mask, d * d, 0)
    sum_square = tl.sum(s, axis=0)
    var = sum_square / N
    rstd = tl.math.rsqrt(var + eps)

    # Store stats if needed (optional pointers)
    if out_mean_ptr is not None: tl.store(out_mean_ptr + pid, m)
    if out_rstd_ptr is not None: tl.store(out_rstd_ptr + pid, rstd)

    w = tl.load(weight_ptr + n_offsets, mask=mask) if weight_ptr is not None else 1.0
    b = tl.load(bias_ptr + n_offsets, mask=mask) if bias_ptr is not None else 0.0
    
    out = (x - m) * rstd * w + b
    tl.store(out_ptr + pid * N + n_offsets, out, mask=mask)

@triton.jit
def layer_norm_loop_kernel(
    in_ptr, out_ptr, weight_ptr, bias_ptr,
    out_mean_ptr, out_rstd_ptr,
    M, N, eps,
    TILE_N: tl.constexpr,
):
    pid = tl.program_id(0)
    # Welford or simple 2-pass? FlagGems uses 2-pass with loop
    
    # 1. Mean
    m = 0.0
    for off in range(0, N, TILE_N):
        cols = off + tl.arange(0, TILE_N)
        mask = cols < N
        val = tl.load(in_ptr + pid * N + cols, mask, other=0.0).to(tl.float32)
        m += tl.sum(val, axis=0)
    m = m / N
    
    # 2. Var
    var = 0.0
    for off in range(0, N, TILE_N):
        cols = off + tl.arange(0, TILE_N)
        mask = cols < N
        val = tl.load(in_ptr + pid * N + cols, mask, other=0.0).to(tl.float32)
        diff = val - m
        var += tl.sum(diff * diff, axis=0)
    var = var / N
    rstd = tl.math.rsqrt(var + eps)
    
    if out_mean_ptr is not None: tl.store(out_mean_ptr + pid, m)
    if out_rstd_ptr is not None: tl.store(out_rstd_ptr + pid, rstd)
    
    # 3. Normalize
    for off in range(0, N, TILE_N):
        cols = off + tl.arange(0, TILE_N)
        mask = cols < N
        val = tl.load(in_ptr + pid * N + cols, mask, other=0.0).to(tl.float32)
        w = tl.load(weight_ptr + cols, mask=mask) if weight_ptr is not None else 1.0
        b = tl.load(bias_ptr + cols, mask=mask) if bias_ptr is not None else 0.0
        
        out = (val - m) * rstd * w + b
        tl.store(out_ptr + pid * N + cols, out, mask=mask)



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import dim_compress, libentry, tl_extra_shim
from flag_gems.utils import triton_lang_extension as tle

pow = tl_extra_shim.pow
logger = logging.getLogger(__name__)


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("vector_norm"), key=["M", "N"])
@triton.jit
def l2_norm_kernel(X, Out, M, N, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Out = Out + pid
    row_mask = pid < M

    _sum = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(X + cols, mask, other=0.0).to(tl.float32)
        _sum += a * a
    sum = tl.sum(_sum, axis=1)

    out = tl.sqrt(sum)[:, None]
    tl.store(Out, out, row_mask)


@libentry()
@triton.jit
def l2_norm_kernel_1(X, Mid, M, BLOCK_SIZE: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    X = X + offset
    Mid = Mid + pid
    mask = offset < M

    x = tl.load(X, mask=mask, other=0.0).to(tl.float32)
    mid = tl.sum(x * x)
    tl.store(Mid, mid)


@libentry()
@triton.jit
def l2_norm_kernel_2(Mid, Out, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    Mid = Mid + offset
    mask = offset < MID_SIZE
    mid = tl.load(Mid, mask=mask, other=0.0).to(tl.float32)
    out = tl.sqrt(tl.sum(mid))
    tl.store(Out, out)


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("vector_norm"), key=["M", "N"])
@triton.jit
def max_norm_kernel(X, Out, M, N, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Out = Out + pid
    row_mask = pid < M

    _max = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(X + cols, mask, other=0.0).to(tl.float32)
        _max = tl.maximum(tl.abs(a), _max)

    max = tl.max(_max, axis=1)
    out = max[:, None]
    tl.store(Out, out, row_mask)


@libentry()
@triton.jit
def max_norm_kernel_1(X, Mid, M, BLOCK_SIZE: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    X = X + offset
    Mid = Mid + pid
    mask = offset < M

    x = tl.load(X, mask=mask, other=0.0).to(tl.float32)
    mid = tl.max(tl.abs(x))
    tl.store(Mid, mid)


@libentry()
@triton.jit
def max_norm_kernel_2(Mid, Out, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    Mid = Mid + offset
    mask = offset < MID_SIZE
    mid = tl.load(Mid, mask=mask, other=0.0).to(tl.float32)
    out = tl.max(mid)
    tl.store(Out, out)


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("vector_norm"), key=["M", "N"])
@triton.jit
def min_norm_kernel(X, Out, M, N, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Out = Out + pid
    row_mask = pid < M

    _min = tl.full([BLOCK_M, BLOCK_N], value=float("inf"), dtype=tl.float32)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(X + cols, mask, other=float("inf")).to(tl.float32)
        _min = tl.minimum(tl.abs(a), _min)

    min = tl.min(_min, axis=1)
    out = min[:, None]
    tl.store(Out, out, row_mask)


@libentry()
@triton.jit
def min_norm_kernel_1(X, Mid, M, BLOCK_SIZE: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    X = X + offset
    Mid = Mid + pid
    mask = offset < M

    x = tl.load(X, mask=mask, other=float("inf")).to(tl.float32)
    mid = tl.min(tl.abs(x))
    tl.store(Mid, mid)


@libentry()
@triton.jit
def min_norm_kernel_2(Mid, Out, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    Mid = Mid + offset
    mask = offset < MID_SIZE
    mid = tl.load(Mid, mask=mask, other=float("inf")).to(tl.float32)
    out = tl.min(mid)
    tl.store(Out, out)


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("vector_norm"), key=["M", "N"])
@triton.jit
def l0_norm_kernel(X, Out, M, N, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):
    pid = tle.program_id(0) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Out = Out + pid
    row_mask = pid < M

    _sum = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(X + cols, mask, other=0).to(tl.float32)
        _sum += tl.where(a != 0, 1, 0)
    sum = tl.sum(_sum, axis=1)
    out = sum[:, None]
    tl.store(Out, out, row_mask)


@libentry()
@triton.jit
def l0_norm_kernel_1(X, Mid, M, BLOCK_SIZE: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    X = X + offset
    Mid = Mid + pid
    mask = offset < M

    x = tl.load(X, mask=mask, other=0.0).to(tl.float32)
    cnt = (x != 0).to(tl.float32)
    mid = tl.sum(cnt)
    tl.store(Mid, mid)


@libentry()
@triton.jit
def l0_norm_kernel_2(Mid, Out, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    Mid = Mid + offset
    mask = offset < MID_SIZE
    mid = tl.load(Mid, mask=mask, other=0.0).to(tl.float32)
    out = tl.sum(mid)
    tl.store(Out, out)


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("vector_norm"), key=["M", "N"])
@triton.jit(do_not_specialize=["ord"])
def v_norm_kernel(X, Out, M, N, ord, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Out = Out + pid
    row_mask = pid < M

    _sum = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(X + cols, mask, other=0.0).to(tl.float32)
        _sum += pow(tl.abs(a), ord)
    sum = tl.sum(_sum, axis=1)
    out = pow(sum, 1 / ord)[:, None]
    tl.store(Out, out, row_mask)


@libentry()
@triton.jit(do_not_specialize=["ord"])
def l1_norm_kernel_1(X, Mid, ord, M, BLOCK_SIZE: tl.constexpr):
    pid = tle.program_id(0).to(tl.int64)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    X = X + offset
    Mid = Mid + pid
    mask = offset < M

    x = tl.load(X, mask=mask, other=0.0).to(tl.float32)
    mid = tl.sum(pow(tl.abs(x), ord))
    tl.store(Mid, mid)


@libentry()
@triton.jit(do_not_specialize=["ord"])
def l1_norm_kernel_2(Mid, Out, ord, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    Mid = Mid + offset
    mask = offset < MID_SIZE
    mid = tl.load(Mid, mask=mask, other=0.0).to(tl.float32)
    out = pow(tl.sum(mid), 1 / ord)
    tl.store(Out, out)


def vector_norm(x, ord=2, dim=None, keepdim=False, dtype=None):
    logger.debug("GEMS VECTOR NORM")
    if dtype is not None:
        dtype = torch.dtype(dtype)
    else:
        dtype = x.dtype
    if dtype not in [torch.float16, torch.float32, torch.bfloat16]:
        raise NotImplementedError(f"vector_norm not implemented for {dtype}")

    with torch_device_fn.device(x.device):
        if (not dim) or len(dim) == x.ndim:
            dim = list(range(x.ndim))
            shape = [1] * x.ndim
            x = dim_compress(x, dim)
            M = x.numel()
            BLOCK_SIZE = triton.next_power_of_2(math.ceil(math.sqrt(M)))
            MID_SIZE = triton.cdiv(M, BLOCK_SIZE)
            BLOCK_MID = triton.next_power_of_2(MID_SIZE)

            mid = torch.empty([MID_SIZE], dtype=dtype, device=x.device)
            out = torch.empty(shape, dtype=dtype, device=x.device)
            if ord == 2:
                l2_norm_kernel_1[(MID_SIZE,)](x, mid, M, BLOCK_SIZE)
                l2_norm_kernel_2[(1,)](mid, out, MID_SIZE, BLOCK_MID)
            elif ord == float("inf"):
                max_norm_kernel_1[(MID_SIZE,)](x, mid, M, BLOCK_SIZE)
                max_norm_kernel_2[(1,)](mid, out, MID_SIZE, BLOCK_MID)
            elif ord == -float("inf"):
                min_norm_kernel_1[(MID_SIZE,)](x, mid, M, BLOCK_SIZE)
                min_norm_kernel_2[(1,)](mid, out, MID_SIZE, BLOCK_MID)
            elif ord == 0:
                l0_norm_kernel_1[(MID_SIZE,)](x, mid, M, BLOCK_SIZE)
                l0_norm_kernel_2[(1,)](mid, out, MID_SIZE, BLOCK_MID)
            else:
                l1_norm_kernel_1[(MID_SIZE,)](x, mid, ord, M, BLOCK_SIZE)
                l1_norm_kernel_2[(1,)](mid, out, ord, MID_SIZE, BLOCK_MID)
        else:
            shape = list(x.shape)
            dim = [d % x.ndim for d in dim]
            x = dim_compress(x, dim)
            N = 1
            for i in dim:
                N *= shape[i]
                shape[i] = 1
            M = x.numel() // N
            out = torch.empty(shape, dtype=dtype, device=x.device)
            grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]),)
            if ord == 2:
                l2_norm_kernel[grid](x, out, M, N)
            elif ord == float("inf"):
                max_norm_kernel[grid](x, out, M, N)
            elif ord == -float("inf"):
                min_norm_kernel[grid](x, out, M, N)
            elif ord == 0:
                l0_norm_kernel[grid](x, out, M, N)
            else:
                v_norm_kernel[grid](x, out, M, N, ord)
    if not keepdim:
        out = out.squeeze(dim=dim)
    return out



import importlib
import logging
import os
from typing import Any, Callable, List, Mapping, Tuple

import torch

from flag_gems.utils.code_cache import code_cache_dir
from flag_gems.utils.code_utils import IndentedBuffer, write_atomic
from flag_gems.utils.shape_utils import (
    MemOverlap,
    has_internal_overlapping,
    restride_dim,
)

logger = logging.getLogger(__name__)


def generate_imports(code: IndentedBuffer) -> IndentedBuffer:
    code.writeline("import torch")
    code.writeline("import triton")
    code.writeline("import triton.language as tl")
    code.newline()
    code.writeline("from flag_gems.utils import libentry")
    code.writeline("from flag_gems import runtime")
    code.writeline("import flag_gems")
    # code.writeline("from flag_gems.utils import triton_lang_extension as tle")
    code.newline()
    code.newline()
    return code


def generate_scatter_kernel(
    rank: int,
    kernel_name: str,
    code: IndentedBuffer,
) -> IndentedBuffer:
    # make the inlined function visible in the context
    code.newline()

    # the autotune function

    code.writeline("def heur_block(args):")
    with code.indent():
        code.writeline("if(flag_gems.vendor_name in ['metax', 'iluvatar']):")
        with code.indent():
            code.writeline("return 256")
        code.writeline("return 128")
    code.newline()
    code.newline()

    code.writeline("def loop_count(args):")
    with code.indent():
        code.writeline("return 4")
    code.newline()
    code.newline()

    # the decorators
    code.writeline("@libentry()")
    code.writeline("@triton.heuristics(")
    with code.indent():
        code.writeline("{")
        with code.indent():
            code.writeline('"BLOCK": heur_block,')
            code.writeline('"LOOP": loop_count,')
        code.writeline("}")
    code.writeline(")")
    inp_stride_vars = ",".join(f"'inp_stride_{i}'" for i in range(rank))
    index_stride_vars = ",".join(f"'index_stride_{i}'" for i in range(rank))
    src_stride_vars = ",".join(f"'src_stride_{i}'" for i in range(rank))
    shape_vars = ",".join(f"'shape_{i}'" for i in range(rank))
    code.writeline(
        f"@triton.jit(do_not_specialize=['N','stride_dim','inp_size_dim',"
        f"{inp_stride_vars},{index_stride_vars},{src_stride_vars},{shape_vars}])"
    )

    # signature
    code.writeline(f"def {kernel_name}(")
    with code.indent():
        if rank > 0:
            code.writeline("src_strided,")
            code.writeline("index,")
            code.writeline("inp,")
            code.writeline("out,")

            stride_args = ", ".join(f"inp_stride_{i}: int" for i in range(rank))
            code.writeline(f"{stride_args}, # stride for inp")

            stride_args = ", ".join(f"index_stride_{i}: int" for i in range(rank))
            code.writeline(f"{stride_args}, # stride for index")

            stride_args = ", ".join(f"src_stride_{i}: int" for i in range(rank))
            code.writeline(f"{stride_args}, # stride for src")

            shape_args = ", ".join(f"shape_{i}: int" for i in range(rank))
            code.writeline(f"{shape_args}, # shape")
            code.writeline("inp_size_dim,")
            code.writeline("stride_dim,")
            code.writeline("N,")
            # reduce options
            code.writeline("IS_ADD: tl.constexpr,")
            code.writeline("IS_MUL: tl.constexpr,")
            code.writeline("BLOCK: tl.constexpr,")
            code.writeline("LOOP: tl.constexpr,")
            code.writeline("INT32_OFFSET: tl.constexpr")

    code.writeline("):")

    # Kernel Code
    with code.indent():
        code.writeline("pid = tl.program_id(0)")
        code.writeline("if not INT32_OFFSET:")
        with code.indent():
            code.writeline("pid = pid.to(tl.int64)")
        code.writeline("offsets = pid * LOOP * BLOCK + tl.arange(0, BLOCK)")

        #   1. Calculate inp_offsets and idx_offsets
        code.writeline("for loop_iter in tl.static_range(LOOP):")
        with code.indent():
            code.writeline("mask = offsets < N")
            code.writeline("cur_idx = offsets")
            code.writeline("if INT32_OFFSET:")
            with code.indent():
                code.writeline("inp_offsets = tl.zeros((BLOCK, ), dtype=tl.int32)")
                code.writeline("idx_offsets = tl.zeros((BLOCK, ), dtype=tl.int32)")
                code.writeline("src_offsets = tl.zeros((BLOCK, ), dtype=tl.int32)")
            code.writeline("else:")
            with code.indent():
                code.writeline("inp_offsets = tl.zeros((BLOCK, ), dtype=tl.int64)")
                code.writeline("idx_offsets = tl.zeros((BLOCK, ), dtype=tl.int64)")
                code.writeline("src_offsets = tl.zeros((BLOCK, ), dtype=tl.int64)")
            for i in range(rank)[::-1]:
                code.writeline("if INT32_OFFSET:")
                with code.indent():
                    code.writeline(f"shape_{i} = shape_{i}.to(tl.int32)")
                    code.writeline(f"inp_stride_{i} = inp_stride_{i}.to(tl.int32)")
                    code.writeline(f"index_stride_{i} = index_stride_{i}.to(tl.int32)")
                    code.writeline(f"src_stride_{i} = src_stride_{i}.to(tl.int32)")
                code.writeline(f"mod = cur_idx % shape_{i}")
                code.writeline(f"inp_offsets += mod * inp_stride_{i}")
                code.writeline(f"idx_offsets += mod * index_stride_{i}")
                code.writeline(f"src_offsets += mod * src_stride_{i}")
                if i != 0:
                    code.writeline(f"cur_idx = cur_idx // shape_{i}")

            #   2. Use offsets to scatter
            code.writeline(
                "cur_src = tl.load(src_strided + src_offsets, mask=mask, other=0)"
            )
            code.writeline(
                "cur_index = tl.load(index + idx_offsets, mask=mask, other=0)"
            )
            code.writeline("if INT32_OFFSET:")
            with code.indent():
                code.writeline("cur_index = cur_index.to(tl.int32)")
                code.writeline("stride_dim = stride_dim.to(tl.int32)")

            code.writeline("dim_offsets = cur_index * stride_dim")
            code.writeline("inp_offsets += dim_offsets")
            code.newline()
            code.writeline("if IS_ADD: ")
            with code.indent():
                code.writeline(
                    "tl.atomic_add(out + inp_offsets, cur_src, mask=mask, sem='relaxed')"
                )
            code.writeline("elif IS_MUL: ")
            with code.indent():
                code.writeline("stop = tl.where(mask, 0, 1).to(tl.int1)")
                code.writeline("block_stop = False")
                code.writeline("while not block_stop:")
                with code.indent():
                    code.writeline
                    code.writeline(
                        "cur_inp = tl.load(out + inp_offsets, mask=mask, other=0)"
                    )
                    code.writeline("res = tl.where(stop, cur_inp, cur_inp * cur_src)")
                    code.writeline(
                        "cas_res = tl.atomic_cas(out + inp_offsets, cur_inp, res, sem='relaxed')"
                    )
                    code.writeline("stop |= cur_inp == cas_res")
                    code.writeline("block_stop = tl.sum(stop.to(tl.int32)) == BLOCK")

            code.writeline("else: ")
            with code.indent():
                code.writeline("tl.store(out + inp_offsets, cur_src, mask=mask)")

            code.writeline("offsets += BLOCK")

    code.newline()
    code.newline()
    return code


def parameter_for_wrapper() -> str:
    # src_strided, index, inp, out, dim, M, N, reduce
    parameters: List[str] = []

    parameters.append("src_strided")
    parameters.append("index")
    parameters.append("inp")
    parameters.append("out")
    parameters.append("dim_size")
    parameters.append("dim_stride")
    parameters.append("N")
    parameters.append("reduce: tl.constexpr=None")
    parameters.append("int32_offset: tl.constexpr=None")

    return ", ".join(parameters)


def generate_destination_passing_wrapper(
    rank: int,
    wrapper_name: str,
    kernel_name: str,
    code: IndentedBuffer,
) -> IndentedBuffer:
    parameters: str = parameter_for_wrapper()
    wrapper_signature: str = f"def {wrapper_name}({parameters}):"
    code.writeline(wrapper_signature)

    with code.indent():
        code.writeline("inp_strides = list(inp.stride())")
        code.writeline("index_strides = index.stride()")
        code.writeline("src_strides = src_strided.stride()")
        code.writeline("index_shapes = list(index.shape)")
        code.writeline("inp_size_dim = dim_size")
        code.writeline("stride_dim = dim_stride")

        code.writeline('IS_ADD = reduce == "add"')
        code.writeline('IS_MUL = reduce == "multiply"')
        code.writeline("int32_offset = int32_offset or True")

        # kernel launch
        code.writeline("grid = lambda meta: (")
        with code.indent():
            code.writeline('triton.cdiv(N, meta["BLOCK"] * meta["LOOP"]), ')
        code.writeline(")")

        kernel_launch: str = f"{kernel_name}[grid]("
        code.writeline(kernel_launch)

        with code.indent():
            code.writeline("src_strided, index, inp, out, ")
            if rank > 0:
                s = ", ".join(f"inp_strides[{i}]" for i in range(rank))
                code.writeline(f"{s},")

                s = ", ".join(f"index_strides[{i}]" for i in range(rank))
                code.writeline(f"{s},")

                s = ", ".join(f"src_strides[{i}]" for i in range(rank))
                code.writeline(f"{s},")

                s = ", ".join(f"index_shapes[{i}]" for i in range(rank))
                code.writeline(f"{s},")

                code.writeline("inp_size_dim,")
                code.writeline("stride_dim,")
                code.writeline("N,")
                # reduce options
                code.writeline("IS_ADD,")
                code.writeline("IS_MUL,")
                code.writeline("INT32_OFFSET=int32_offset,")
        code.writeline(")")
        code.writeline("return out")

    return code


def generate_code(
    inputs: Tuple[Any],
    wrapper_name: str,
    kernel_name: str,
    code: IndentedBuffer,
) -> IndentedBuffer:
    # inputs: [src_strided, index, inp, out, dim, M, N, reduce]
    shape = inputs[1].shape
    rank = len(shape)

    code = generate_imports(code)
    code = generate_scatter_kernel(rank, kernel_name, code)
    code = generate_destination_passing_wrapper(rank, wrapper_name, kernel_name, code)
    return code


class ScatterFunction:
    def __init__(self):
        self.pid = os.getpid()
        self.overloads: Mapping[str, Callable] = {}

    def __call__(self, *args, **kwargs):
        key = f"{self.arg_key(*args)}"
        if key in self.overloads:
            overload = self.overloads[key]
        else:
            code = IndentedBuffer()
            code = generate_code(
                args,
                "_scatter_wrapper",
                "_scatter_jit_function",
                code,
            )

            file_name = f"scatter_rank_{key}.py"
            file_path = code_cache_dir() / file_name
            write_atomic(file_path, code.getvalue())

            # load
            spec = importlib.util.spec_from_file_location(
                f"_gen_module_rank_{key}",
                file_path,
            )

            m = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(m)
            overload = getattr(m, "_scatter_wrapper")
            self.overloads[key] = overload

        return overload(*args, **kwargs)

    def arg_key(self, *args):
        tensors = [item for item in args if torch.is_tensor(item)]
        max_rank = max(item.ndim for item in tensors)
        return max_rank


_scatter_func = ScatterFunction()


def scatter(inp, dim, index, src, reduce=None):
    logger.debug("GEMS SCATTER")
    out = inp.clone()

    if reduce is not None:
        assert inp.dtype not in (
            torch.bfloat16,
        ), "Unsupported operation: reduce scatter bfloat tensors."

    if has_internal_overlapping(out) == MemOverlap.Yes:
        out = out.contiguous()

    src_strided = src.as_strided(index.shape, src.stride())
    inp_restrided = restride_dim(inp, dim, index.shape)
    dim_size = inp.size(dim)
    dim_stride = inp.stride(dim)
    N = index.numel()

    int32_size_dim = lambda x: x.stride(dim) * x.size(dim) < 2**32
    use_int32_offset = all(map(int32_size_dim, (inp, index, src)))
    _scatter_func(
        src_strided,
        index,
        inp_restrided,
        out,
        dim_size,
        dim_stride,
        N,
        reduce,
        int32_offset=use_int32_offset,
    )

    return out


def scatter_(inp, dim, index, src, reduce=None):
    logger.debug("GEMS SCATTER_")
    out = inp

    if reduce is not None:
        assert inp.dtype not in (
            torch.bfloat16,
        ), "Unsupported operation: reduce scatter bfloat tensors."

    assert (
        has_internal_overlapping(out) != MemOverlap.Yes
    ), "Unsupported operation: trying to inplace write to an internally overlapping tensor."

    src_restrided = src.as_strided(index.shape, src.stride())
    inp_restrided = restride_dim(inp, dim, index.shape)
    dim_size = inp.size(dim)
    dim_stride = inp.stride(dim)
    N = index.numel()

    int32_size_dim = lambda x: x.stride(dim) * x.size(dim) < 2**32
    use_int32_offset = all(map(int32_size_dim, (inp, index, src)))
    _scatter_func(
        src_restrided,
        index,
        inp_restrided,
        out,
        dim_size,
        dim_stride,
        N,
        reduce,
        int32_offset=use_int32_offset,
    )

    return inp



# Batch MatMul - Pure Triton
# Source: FlagGems / NeuroBrix Robust
# NeuroBrix - NVIDIA Common

import torch
import triton
import triton.language as tl

@triton.jit
def _bmm_kernel(
    a_ptr, b_ptr, c_ptr,
    B, M, N, K,
    stride_ab, stride_am, stride_ak,
    stride_bb, stride_bk, stride_bn,
    stride_cb, stride_cm, stride_cn,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
):
    """Batched matmul kernel - pure Triton with explicit strides."""
    pid_batch = tl.program_id(2)
    pid = tl.program_id(0)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)

    pid_m = pid // num_pid_n
    pid_n = pid % num_pid_n

    offs_am = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_bn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    offs_k = tl.arange(0, BLOCK_SIZE_K)

    a_ptrs = a_ptr + pid_batch * stride_ab + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + pid_batch * stride_bb + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)

    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)

    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        a = tl.load(a_ptrs, mask=(offs_am[:, None] < M) & (offs_k[None, :] < K - k * BLOCK_SIZE_K), other=0.0)
        b = tl.load(b_ptrs, mask=(offs_k[:, None] < K - k * BLOCK_SIZE_K) & (offs_bn[None, :] < N), other=0.0)
        accumulator += tl.dot(a, b, allow_tf32=False)
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk

    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + pid_batch * stride_cb + stride_cm * offs_cm[:, None] + stride_cn * offs_cn[None, :]
    c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)
    # Clamp to fp16 range before store to prevent inf
    accumulator = tl.minimum(tl.maximum(accumulator, -65000.0), 65000.0)
    tl.store(c_ptrs, accumulator.to(c_ptr.dtype.element_ty), mask=c_mask)



"""
NeuroBrix Execution Strategies

Strategy classes that handle component execution based on Prism allocation decisions.
The executor delegates to these classes — NO placement decisions in executor.

Strategy Hierarchy (granularity):
  Single GPU → Component Placement → Pipeline Parallel (layer) → Block Scatter → Weight Sharding

Every strategy listed here is a first-class citizen — NeuroBrix is universal
and must handle any hardware combination. Prism scores ALL strategies and
selects the best viable one for the given hardware profile.

Strategies:
- SingleGPUStrategy: All components on one GPU
- ComponentPlacementStrategy: Whole components distributed across GPUs
- ComponentPlacementLazyStrategy: Component placement with lazy weight swap
- PipelineParallelStrategy: Per-layer sequential fill across GPUs (like Accelerate)
- BlockScatterStrategy: Block-level best-fit distribution across GPUs
- WeightShardingStrategy: Weight-file-level round-robin across GPUs
- LazySequentialStrategy: One component at a time on largest GPU
- Zero3Strategy: CPU offload with GPU compute streaming
"""

from .base import ExecutionStrategy, StrategyContext
from .single_gpu import SingleGPUStrategy
from .component_placement import ComponentPlacementStrategy, ComponentPlacementLazyStrategy
from .pipeline_parallel import PipelineParallelStrategy
from .block_scatter import BlockScatterStrategy
from .weight_sharding import WeightShardingStrategy
from .lazy_sequential import LazySequentialStrategy
from .zero3 import Zero3Strategy


# =============================================================================
# STRATEGY REGISTRY
# =============================================================================
# Every strategy name that Prism can emit MUST have a registry entry.
# ZERO FALLBACK: get_strategy() crashes on unknown names.
#
# Format: "strategy_name" → StrategyClass
# The solver uses these exact string names in ExecutionPlan.strategy.
# =============================================================================

STRATEGY_REGISTRY = {
    # === Single Device ===
    "single_gpu": SingleGPUStrategy,
    "single_gpu_lifecycle": SingleGPUStrategy,

    # === Component Placement (whole-component distribution) ===
    "component_placement": ComponentPlacementStrategy,
    "component_placement_lazy": ComponentPlacementLazyStrategy,

    # === Pipeline Parallel (per-layer sequential fill) ===
    "pipeline_parallel": PipelineParallelStrategy,

    # === Block Scatter (block-level best-fit distribution) ===
    "block_scatter": BlockScatterStrategy,

    # === Weight Sharding (weight-file round-robin) ===
    "weight_sharding": WeightShardingStrategy,

    # === Sequential / Offload ===
    "lazy_sequential": LazySequentialStrategy,
    "zero3": Zero3Strategy,
}


def get_strategy(strategy_name: str, context: StrategyContext) -> ExecutionStrategy:
    """
    Get strategy instance based on Prism's decision.

    ZERO FALLBACK: Crash if unknown strategy
    """
    if strategy_name not in STRATEGY_REGISTRY:
        raise RuntimeError(
            f"ZERO FALLBACK: Unknown strategy '{strategy_name}'. "
            f"Available: {sorted(STRATEGY_REGISTRY.keys())}"
        )

    strategy_class = STRATEGY_REGISTRY[strategy_name]
    return strategy_class(context, strategy_name)


__all__ = [
    "ExecutionStrategy",
    "StrategyContext",
    "SingleGPUStrategy",
    "ComponentPlacementStrategy",
    "ComponentPlacementLazyStrategy",
    "PipelineParallelStrategy",
    "BlockScatterStrategy",
    "WeightShardingStrategy",
    "LazySequentialStrategy",
    "Zero3Strategy",
    "get_strategy",
    "STRATEGY_REGISTRY",
]



import logging
import math

import torch
import triton
import triton.language as tl
import triton.runtime.driver as driver

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


def get_npu_properties():
    device = torch.npu.current_device()
    return driver.active.utils.get_device_properties(device)


@triton.jit
def resolve_neg_kernel(
    inp,
    out,
    data_len: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    MAX_DATA_SIZE: tl.constexpr,
):
    pid = tl.program_id(0)
    iter_num = tl.cdiv(BLOCK_SIZE, MAX_DATA_SIZE)

    for idx in tl.range(0, iter_num):
        offsets = pid * BLOCK_SIZE + idx * MAX_DATA_SIZE + tl.arange(0, MAX_DATA_SIZE)
        mask = offsets < data_len
        inp_val = tl.load(inp + offsets, mask=mask)
        out_val = -inp_val
        tl.store(out + offsets, out_val, mask=mask)


def resolve_neg(A: torch.Tensor):
    logger.debug("GEMS_ASCEND RESOLVE_NEG")

    if A.is_neg():
        data_len = A.numel()
        out = torch.empty(A.numel(), dtype=A.dtype, device=A.device)

        CORE_NUM = get_npu_properties()["num_vectorcore"]
        BLOCK_SIZE = math.ceil(data_len / CORE_NUM)
        MAX_DATA_SIZE = 20 * 1024

        grid = lambda meta: (triton.cdiv(data_len, meta["BLOCK_SIZE"]),)
        resolve_neg_kernel[grid](
            A,
            out,
            data_len,
            BLOCK_SIZE,
            MAX_DATA_SIZE,
        )
        out = out.view(A.shape)
        return out
    else:
        return A



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

import warnings

import torch

from fla.modules.l2norm import l2norm_bwd, l2norm_fwd
from fla.ops.common.chunk_delta_h import chunk_gated_delta_rule_bwd_dhu, chunk_gated_delta_rule_fwd_h
from fla.ops.common.chunk_o import chunk_bwd_dqkwg, chunk_bwd_dv_local, chunk_fwd_o
from fla.ops.delta_rule.wy_fast import prepare_wy_repr_bwd, prepare_wy_repr_fwd, recompute_w_u_fwd
from fla.utils import autocast_custom_bwd, autocast_custom_fwd, input_guard


def chunk_delta_rule_fwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    beta: torch.Tensor,
    scale: float,
    initial_state: torch.Tensor,
    output_final_state: bool,
    cu_seqlens: torch.LongTensor | None = None,
):
    # obtain WY representation. u is actually the new v.
    w, u, A = prepare_wy_repr_fwd(
        k=k,
        v=v,
        beta=beta,
        cu_seqlens=cu_seqlens,
    )
    h, v_new, final_state = chunk_gated_delta_rule_fwd_h(
        k=k,
        w=w,
        u=u,
        g=None,
        initial_state=initial_state,
        output_final_state=output_final_state,
        cu_seqlens=cu_seqlens,
    )

    o = chunk_fwd_o(
        q=q,
        k=k,
        v=v_new,
        h=h,
        g=None,
        scale=scale,
        cu_seqlens=cu_seqlens,
    )
    return o, A, final_state


def chunk_delta_rule_bwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    beta: torch.Tensor,
    A: torch.Tensor,
    scale: float,
    initial_state: torch.Tensor,
    do: torch.Tensor,
    dht: torch.Tensor,
    cu_seqlens: torch.LongTensor | None = None,
):
    w, u = recompute_w_u_fwd(
        k=k,
        v=v,
        beta=beta,
        A=A,
        cu_seqlens=cu_seqlens,
    )
    h, v_new, _ = chunk_gated_delta_rule_fwd_h(
        k=k,
        w=w,
        u=u,
        g=None,
        initial_state=initial_state,
        output_final_state=False,
        cu_seqlens=cu_seqlens,
    )
    dv = chunk_bwd_dv_local(
        q=q,
        k=k,
        do=do,
        g=None,
        scale=scale,
        cu_seqlens=cu_seqlens,
    )
    dh, dh0, dv = chunk_gated_delta_rule_bwd_dhu(
        q=q,
        k=k,
        w=w,
        g=None,
        h0=initial_state,
        dht=dht,
        do=do,
        dv=dv,
        scale=scale,
        cu_seqlens=cu_seqlens,
    )
    dq, dk, dw, _ = chunk_bwd_dqkwg(
        q=q,
        k=k,
        v=v_new,
        h=h,
        w=w,
        dv=dv,
        do=do,
        dh=dh,
        g=None,
        scale=scale,
        cu_seqlens=cu_seqlens,
    )
    dk2, dv, db = prepare_wy_repr_bwd(
        k=k,
        v=v,
        beta=beta,
        A=A,
        dw=dw,
        du=dv,
        cu_seqlens=cu_seqlens,
    )
    dk.add_(dk2)
    return dq, dk, dv, db, dh0


class ChunkDeltaRuleFunction(torch.autograd.Function):

    @staticmethod
    @input_guard
    @autocast_custom_fwd
    def forward(
        ctx,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        beta: torch.Tensor,
        scale: float,
        initial_state: torch.Tensor,
        output_final_state: bool,
        use_qk_l2norm_in_kernel: bool = False,
        cu_seqlens: torch.LongTensor | None = None,
    ):
        if use_qk_l2norm_in_kernel:
            q, q_rstd = l2norm_fwd(q)
            k, k_rstd = l2norm_fwd(k)
        else:
            q_rstd, k_rstd = None, None

        o, A, final_state = chunk_delta_rule_fwd(
            q=q,
            k=k,
            v=v,
            beta=beta,
            scale=scale,
            initial_state=initial_state,
            output_final_state=output_final_state,
            cu_seqlens=cu_seqlens,
        )
        ctx.save_for_backward(q, q_rstd, k, k_rstd, v, beta, A, initial_state)
        ctx.scale = scale
        ctx.cu_seqlens = cu_seqlens
        ctx.use_qk_l2norm_in_kernel = use_qk_l2norm_in_kernel
        return o.to(q.dtype), final_state

    @staticmethod
    @input_guard
    @autocast_custom_bwd
    def backward(
        ctx,
        do: torch.Tensor,
        dht: torch.Tensor,
    ):
        q, q_rstd, k, k_rstd, v, beta, A, initial_state = ctx.saved_tensors

        dq, dk, dv, db, dh0 = chunk_delta_rule_bwd(
            q=q,
            k=k,
            v=v,
            beta=beta,
            A=A,
            scale=ctx.scale,
            initial_state=initial_state,
            do=do,
            dht=dht,
            cu_seqlens=ctx.cu_seqlens,
        )
        if ctx.use_qk_l2norm_in_kernel:
            dq = l2norm_bwd(q, q_rstd, dq)
            dk = l2norm_bwd(k, k_rstd, dk)
        return dq.to(q.dtype), dk.to(k.dtype), dv.to(v.dtype), db.to(beta.dtype), None, dh0, None, None, None, None, None


@torch.compiler.disable
def chunk_delta_rule(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    beta: torch.Tensor,
    scale: float = None,
    initial_state: torch.Tensor = None,
    output_final_state: bool = False,
    use_qk_l2norm_in_kernel: bool = False,
    cu_seqlens: torch.LongTensor | None = None,
    head_first: bool = False,
):
    r"""
    Args:
        q (torch.Tensor):
            queries of shape `[B, T, H, K]`.
        k (torch.Tensor):
            keys of shape `[B, T, H, K]`.
        v (torch.Tensor):
            values of shape `[B, T, H, V]`.
        beta (torch.Tensor):
            betas of shape `[B, T, H]`.
        scale (Optional[float]):
            Scale factor for the RetNet attention scores.
            If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
        initial_state (Optional[torch.Tensor]):
            Initial state of shape `[N, H, K, V]` for `N` input sequences.
            For equal-length input sequences, `N` equals the batch size `B`.
            Default: `None`.
        output_final_state (Optional[bool]):
            Whether to output the final state of shape `[N, H, K, V]`. Default: `False`.
        use_qk_l2norm_in_kernel (Optional[bool]):
            Whether to use qk l2norm within the kernel for saving GPU memory.
            Default: `False`.
        cu_seqlens (torch.LongTensor):
            Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
            consistent with the FlashAttention API.
        head_first (Optional[bool]):
            Whether the inputs are in the head-first format. Default: `False`.
            This argument has been deprecated.

    Returns:
        o (torch.Tensor):
            Outputs of shape `[B, T, H, V]`.
        final_state (torch.Tensor):
            Final state of shape `[N, H, K, V]` if `output_final_state=True` else `None`.

    Examples::
        >>> import torch
        >>> import torch.nn.functional as F
        >>> from einops import rearrange
        >>> from fla.ops.delta_rule import chunk_delta_rule
        # inputs with equal lengths
        >>> B, T, H, K, V = 4, 2048, 4, 512, 512
        >>> q = torch.randn(B, T, H, K, dtype=torch.bfloat16, device='cuda')
        >>> k = F.normalize(torch.randn(B, T, H, K, dtype=torch.bfloat16, device='cuda'), p=2, dim=-1)
        >>> v = torch.randn(B, T, H, V, dtype=torch.bfloat16, device='cuda')
        >>> beta = torch.rand(B, T, H, dtype=torch.bfloat16, device='cuda').sigmoid()
        >>> h0 = torch.randn(B, H, K, V, dtype=torch.bfloat16, device='cuda')
        >>> o, ht = chunk_delta_rule(
            q, k, v, beta,
            initial_state=h0,
            output_final_state=True
        )
        # for variable-length inputs, the batch size `B` is expected to be 1 and `cu_seqlens` is required
        >>> q, k, v, beta = map(lambda x: rearrange(x, 'b t ... -> 1 (b t) ...'), (q, k, v, beta))
        # for a batch with 4 sequences, `cu_seqlens` with 5 start/end positions are expected
        >>> cu_seqlens = q.new_tensor([0, 2048, 4096, 6144, 8192], dtype=torch.long)
        >>> o, ht = chunk_delta_rule(
            q, k, v, beta,
            initial_state=h0,
            output_final_state=True,
            cu_seqlens=cu_seqlens
        )
    """
    assert q.dtype == k.dtype == v.dtype
    assert q.dtype != torch.float32, "ChunkDeltaRuleFunction does not support float32. Please use bfloat16."
    assert len(beta.shape) == 3, "beta must be of shape (batch size, num of head, seq len)."

    if head_first:
        raise DeprecationWarning(
            "head_first is deprecated and will be removed in a future version. "
            "Please use head_first=False for now instead.",
        )
    if not head_first and q.shape[1] < q.shape[2]:
        warnings.warn(
            f"Input tensor shape suggests potential format mismatch: seq_len ({q.shape[1]}) < num_heads ({q.shape[2]}). "
            "This may indicate the inputs were passed in head-first format [B, H, T, ...] "
            "when head_first=False was specified. "
            "Please verify your input tensor format matches the expected shape [B, T, H, ...].",
        )
    if cu_seqlens is not None:
        if q.shape[0] != 1:
            raise ValueError(
                f"The batch size is expected to be 1 rather than {q.shape[0]} when using `cu_seqlens`."
                f"Please flatten variable-length inputs before processing.",
            )
        if initial_state is not None and initial_state.shape[0] != len(cu_seqlens) - 1:
            raise ValueError(
                f"The number of initial states is expected to be equal to the number of input sequences, "
                f"i.e., {len(cu_seqlens) - 1} rather than {initial_state.shape[0]}.",
            )
    scale = k.shape[-1] ** -0.5 if scale is None else scale
    o, final_state = ChunkDeltaRuleFunction.apply(
        q,
        k,
        v,
        beta,
        scale,
        initial_state,
        output_final_state,
        use_qk_l2norm_in_kernel,
        cu_seqlens,
    )
    return o, final_state



"""
Static Graph Flow Handler

ZERO SEMANTIC: Executes all components in topological order.
ZERO HARDCODE: All configuration from NBX container.

Handles models without iteration - single pass through all components.
"""

import gc
import torch
from typing import Any, Callable, Dict, Optional

from .base import FlowHandler, FlowContext, register_flow


@register_flow("static_graph")
class StaticGraphHandler(FlowHandler):
    """
    Flow handler for static graph execution (no loop).

    Executes all components in topological order without iteration.
    Used for simple models that don't need denoising loops.

    ZERO SEMANTIC: No domain knowledge.
    ZERO HARDCODE: Component order from topology.
    """

    def __init__(
        self,
        ctx: FlowContext,
        execute_component_fn: Callable[[str, str, Optional[torch.Tensor]], Any]
    ):
        """
        Initialize static graph handler.

        Args:
            ctx: FlowContext with all shared state
            execute_component_fn: Function to execute a component
        """
        super().__init__(ctx)
        self._execute_component = execute_component_fn

    def execute(self) -> Dict[str, Any]:
        """
        Execute static graph (no loop).

        Executes all components in order: pre_loop + loop.components + post_loop

        Returns:
            Dict of resolved variables/outputs
        """
        flow = self.ctx.pkg.topology.get("flow", {})

        # Execute all components in topological order
        all_components = (
            flow.get("pre_loop", []) +
            flow.get("loop", {}).get("components", []) +
            flow.get("post_loop", [])
        )

        for comp_name in all_components:
            self._execute_component(comp_name, "static", None)

            # Unload after each component
            self._unload_component(comp_name)

        return self.ctx.variable_resolver.resolve_all()

    def _unload_component(self, comp_name: str) -> None:
        """Unload component weights and clear memory (skip in serve mode)."""
        if self.ctx.persistent_mode:
            return
        executor = self.ctx.executors.get(comp_name)
        if executor:
            executor.unload_weights()
        gc.collect()
        device_empty_cache(self.ctx.primary_device)



# core/dtype/__init__.py
"""
NeuroBrix Dtype Module - Single Source of Truth

Consolidates all dtype-related code:
1. config.py - DTYPE_MAP, BYTES_MAP, HARDWARE_SUPPORT
2. converter.py - safe_dtype_convert(), calculate_dtype_multiplier()
3. engine.py - DtypeEngine (single dtype decision engine)

Previously duplicated across:
- core/prism/common/dtype_resolver.py
- core/prism/common/allocation.py
- core/runtime/weight_loader.py

ZERO HARDCODE: Import from here for all dtype operations.
"""

# Config - constants and mappings
from neurobrix.core.dtype.config import (
    DTYPE_MAP,
    DTYPE_TO_STR,
    BYTES_MAP,
    HARDWARE_DTYPE_SUPPORT,
    get_dtype_bytes,
    get_torch_dtype,
    dtype_to_str,
    architecture_supports_dtype,
    parse_dtype,
    strip_aten_prefix,
)

# Converter - safe conversion functions
from neurobrix.core.dtype.converter import (
    safe_dtype_convert,
    safe_dtype_convert_dict,
    calculate_dtype_multiplier,
    resolve_safe_fallback,
)

# DtypeEngine - single entry point for all dtype decisions
from neurobrix.core.dtype.engine import DtypeEngine

__all__ = [
    # Config
    "DTYPE_MAP",
    "DTYPE_TO_STR",
    "BYTES_MAP",
    "HARDWARE_DTYPE_SUPPORT",
    "get_dtype_bytes",
    "get_torch_dtype",
    "dtype_to_str",
    "architecture_supports_dtype",
    "parse_dtype",
    "strip_aten_prefix",
    # Converter
    "safe_dtype_convert",
    "safe_dtype_convert_dict",
    "calculate_dtype_multiplier",
    "resolve_safe_fallback",
    # DtypeEngine
    "DtypeEngine",
]



import logging

import triton

from flag_gems.utils import pointwise_dynamic

logger = logging.getLogger(__name__)


@pointwise_dynamic(promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def bitwise_and_func(x, y):
    return x & y


def bitwise_and_tensor(A, B):
    logger.debug("GEMS BITWISE AND")
    return bitwise_and_func(A, B)


def bitwise_and_tensor_(A, B):
    logger.debug("GEMS BITWISE AND_")
    return bitwise_and_func(A, B, out0=A)


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def bitwise_and_func_scalar(x, y):
    return x & y


def bitwise_and_scalar(A, B):
    logger.debug("GEMS BITWISE AND SCALAR")
    return bitwise_and_func_scalar(A, B)


def bitwise_and_scalar_(A, B):
    logger.debug("GEMS BITWISE AND_ SCALAR")
    return bitwise_and_func_scalar(A, B, out0=A)


def bitwise_and_scalar_tensor(A, B):
    logger.debug("GEMS BITWISE AND SCALAR TENSOR")
    return bitwise_and_func_scalar(B, A)



"""
Unified Kernel Resolver - NeuroBrix.

Combines ATen classification with cascade kernel resolution.
Single entry point for all kernel lookups.

Resolution Flow:
1. Classify op (TRITON/METADATA)
2. Map to kernel file name
3. Cascade through tiers: arch-specific → common
4. Return kernel function or raise if not found

ZERO FALLBACK: Missing kernel = explicit error with guidance.
"""

import os
import sys
import importlib
from pathlib import Path
from typing import Callable, Optional, List, Dict, Any, Union

from .registry import KERNEL_REGISTRY
from .classification import OpExecution, get_execution_type, ATEN_CLASSIFICATION
from .mapping import get_kernel_op_name, ATEN_TO_KERNEL
from .exceptions import KernelNotFoundError


# =============================================================================
# PERFORMANCE CACHES - Critical for runtime speed
# =============================================================================

# Cache: (family, vendor, arch, kernel_op) -> module
_MODULE_CACHE: Dict[tuple, Any] = {}

# Cache: op_type_full -> (execution_type, kernel_op_name)
_OP_INFO_CACHE: Dict[str, tuple] = {}


# =============================================================================
# Architecture → Tier Cascade Mapping
# =============================================================================

NVIDIA_TIER_CASCADE = {
    "hopper": ["hopper", "ampere", "volta", "common"],
    "ada": ["ada", "ampere", "volta", "common"],
    "ampere": ["ampere", "volta", "common"],
    "turing": ["turing", "volta", "common"],
    "volta": ["volta", "common"],
    "common": ["common"],
}

AMD_TIER_CASCADE = {
    "cdna3": ["cdna3", "cdna2", "cdna", "common"],
    "cdna2": ["cdna2", "cdna", "common"],
    "cdna": ["cdna", "common"],
    "rdna3": ["rdna3", "rdna2", "common"],
    "rdna2": ["rdna2", "common"],
    "common": ["common"],
}


def _get_tier_cascade(vendor: str, arch: str) -> List[str]:
    """Get tier cascade for vendor/architecture."""
    if vendor == "nvidia":
        return NVIDIA_TIER_CASCADE.get(arch.lower(), ["common"])
    elif vendor == "amd":
        return AMD_TIER_CASCADE.get(arch.lower(), ["common"])
    else:
        return ["common"]


# =============================================================================
# Auto-Discovery (import all kernel modules to trigger @register_kernel)
# =============================================================================

_kernels_loaded = False


def _ensure_kernels_loaded():
    """Import all kernel modules to trigger @register_kernel decorators."""
    global _kernels_loaded
    if _kernels_loaded:
        return

    kernels_dir = Path(__file__).parent
    ops_dir = kernels_dir / "ops"

    if not ops_dir.exists():
        return

    for root, dirs, files in os.walk(ops_dir):
        # Skip __pycache__
        if "__pycache__" in dirs:
            dirs.remove("__pycache__")

        for file in files:
            if file.endswith(".py") and not file.startswith("_"):
                file_path = Path(root) / file
                try:
                    # Convert path to module: kernels/ops/image/nvidia/common/add.py
                    # → kernels.ops.image.nvidia.common.add
                    rel_path = file_path.relative_to(kernels_dir)
                    module_parts = list(rel_path.with_suffix("").parts)
                    module_name = "neurobrix.kernels." + ".".join(module_parts)
                    importlib.import_module(module_name)
                except Exception:
                    # Log but dont crash - maybe syntax error in one kernel
                    pass

    _kernels_loaded = True


# =============================================================================
# Public API
# =============================================================================

def get_kernel(
    op_type: str,
    family: str = "image",
    vendor: str = "nvidia",
    arch: str = "volta",
    hw: Dict[str, str] = None,
) -> Optional[Callable]:
    """
    Get kernel function for an ATen operation.

    Args:
        op_type: ATen op name (e.g., "aten::addmm" or "addmm")
        family: Model family (image, llm, video, audio)
        vendor: Hardware vendor (nvidia, amd)
        arch: Architecture (volta, ampere, hopper, etc.)
        hw: Optional hardware dict with vendor/architecture keys

    Returns:
        Kernel function or None if metadata op

    Raises:
        KernelNotFoundError: If op unknown or kernel not found (ZERO FALLBACK)
    """
    _ensure_kernels_loaded()

    # Extract from hw dict if provided
    if hw is not None:
        vendor = hw.get("vendor", vendor)
        arch = hw.get("architecture", arch)

    # Normalize op name
    if not op_type.startswith("aten::"):
        op_type_full = f"aten::{op_type}"
    else:
        op_type_full = op_type

    # 1. Check if known ATen op
    if op_type_full in ATEN_CLASSIFICATION:
        # Classify operation
        exec_type = get_execution_type(op_type_full)

        # Metadata ops - no kernel needed
        if exec_type == OpExecution.METADATA:
            return None

        # Get kernel operation name
        kernel_op = get_kernel_op_name(op_type_full)
        if kernel_op is None:
            return None
    else:
        # Direct kernel name (not ATen op)
        kernel_op = op_type.replace("aten::", "")

    # 2. Try registry with cascade
    # Return PURE KERNEL ONLY - adapter handles all translation
    tiers = _get_tier_cascade(vendor, arch)
    for tier in tiers:
        key = (family, vendor, tier, kernel_op)
        if key in KERNEL_REGISTRY:
            return KERNEL_REGISTRY[key].impl

    # 3. Try dynamic import with cascade (for auto-registration)
    for tier in tiers:
        try:
            module_path = f"neurobrix.kernels.ops.{family}.{vendor}.{tier}.{kernel_op}"
            importlib.import_module(module_path)

            # After import, check registry again
            key = (family, vendor, tier, kernel_op)
            if key in KERNEL_REGISTRY:
                return KERNEL_REGISTRY[key].impl

        except ImportError:
            continue

    # 4. ZERO FALLBACK - explicit error
    available_ops = _list_available_ops(family, vendor)
    tiers_str = " → ".join(tiers)

    raise KernelNotFoundError(
        f"No kernel found for {op_type_full}\n"
        f"Context: {family}/{vendor}/{arch}\n"
        f"Tiers searched: {tiers_str}"
    )


def get_kernel_module(
    op_type: str,
    family: str = "image",
    vendor: str = "nvidia",
    arch: str = "volta",
) -> Optional[Any]:
    """
    Get the full kernel module for an op.
    Used by adapter to access JIT functions directly.

    PERFORMANCE: Uses aggressive caching to minimize per-op overhead.
    """
    # Fast path: check op info cache first
    if not op_type.startswith("aten::"):
        op_type_full = f"aten::{op_type}"
    else:
        op_type_full = op_type

    # Get kernel op name (cached lookup)
    if op_type_full in _OP_INFO_CACHE:
        _, kernel_op = _OP_INFO_CACHE[op_type_full]
    else:
        kernel_op = get_kernel_op_name(op_type_full)
        if kernel_op is not None:
            kernel_op = kernel_op.replace("aten::", "")
        exec_type = get_execution_type(op_type_full) if op_type_full in ATEN_CLASSIFICATION else None
        _OP_INFO_CACHE[op_type_full] = (exec_type, kernel_op)

    if kernel_op is None:
        return None

    # Check module cache (fastest path)
    cache_key = (family, vendor, arch, kernel_op)
    if cache_key in _MODULE_CACHE:
        return _MODULE_CACHE[cache_key]

    # Cascade through tiers
    tiers = _get_tier_cascade(vendor, arch)
    for tier in tiers:
        module_path = f"neurobrix.kernels.ops.{family}.{vendor}.{tier}.{kernel_op}"

        # Check sys.modules first
        if module_path in sys.modules:
            module = sys.modules[module_path]
            _MODULE_CACHE[cache_key] = module
            return module

        # Try import
        try:
            module = importlib.import_module(module_path)
            _MODULE_CACHE[cache_key] = module
            return module
        except ImportError:
            continue

    # Cache None to avoid repeated failed lookups
    _MODULE_CACHE[cache_key] = None
    return None


def run_op(
    op_type: str,
    inputs: List,
    outputs: List = None,
    attrs: Dict = None,
    family: str = "image",
    vendor: str = "nvidia",
    arch: str = "volta",
    hw: Dict[str, str] = None,
) -> Any:
    """
    Resolve and execute kernel in one call.

    For METADATA ops, returns None (caller handles via PyTorch).
    """
    kernel = get_kernel(op_type, family, vendor, arch, hw)

    if kernel is None:
        # Metadata op - caller should handle
        return None

    # Support both old (inputs, outputs, attrs) and new (*tensors) calling conventions
    if outputs is not None or attrs is not None:
        return kernel(inputs, outputs or [], attrs or {})
    else:
        # inputs is actually a list of tensors
        return kernel(*inputs)


def list_kernels(
    family: str = None,
    vendor: str = None,
    tier: str = None,
) -> List[tuple]:
    """List all registered kernels with optional filters."""
    _ensure_kernels_loaded()

    keys = list(KERNEL_REGISTRY.keys())

    if family:
        keys = [k for k in keys if k[0] == family]
    if vendor:
        keys = [k for k in keys if k[1] == vendor]
    if tier:
        keys = [k for k in keys if k[2] == tier]

    return keys


def _list_available_ops(family: str, vendor: str) -> List[str]:
    """List available ops for a family/vendor."""
    keys = list_kernels(family=family, vendor=vendor)
    ops = sorted(set(k[3] for k in keys))
    return ops if ops else ["none"]



import logging
import os

import torch
import triton
import triton.language as tl

# from flag_gems import runtime
from flag_gems.utils import dim_compress, libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@triton.jit
def count_nonzero_kernel_1(x_ptr, out_ptr, numel, BLOCK_SIZE: tl.constexpr):
    pid = tle.program_id(0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < numel
    x = tl.load(x_ptr + offsets, mask=mask, other=0)
    is_nonzero = (x != 0).to(tl.int64)
    nonzero_count = tl.sum(is_nonzero, axis=0)
    tl.atomic_add(out_ptr, nonzero_count)


"""***************************** TROTITON XPU KERNEL *****************************"""


@libentry()
@triton.jit
def count_nonzero_kernel_1_part0_xpu(x_ptr, out_ptr, numel, BLOCK_SIZE_0: tl.constexpr):
    pid = tle.program_id(0)
    block_start = pid * BLOCK_SIZE_0
    offsets = block_start + tl.arange(0, BLOCK_SIZE_0)
    mask = offsets < numel
    x = tl.load(x_ptr + offsets, mask=mask, other=0)
    is_nonzero = (x != 0).to(tl.int64)
    nonzero_count = tl.sum(is_nonzero, axis=0)
    tl.store(out_ptr + pid, nonzero_count)


@libentry()
@triton.jit
def count_nonzero_kernel_1_part1_xpu(x_ptr, out_ptr, numel, BLOCK_SIZE_1: tl.constexpr):
    offsets = tl.arange(0, BLOCK_SIZE_1)
    mask = offsets < numel
    x = tl.load(x_ptr + offsets, mask=mask, other=0)
    nonzero_count = tl.sum(x, axis=0)
    tl.store(out_ptr, nonzero_count)


"""***************************** TROTITON XPU KERNEL *****************************"""


def heur_block_size(args):
    return triton.next_power_of_2(triton.cdiv(args["numel"], 12))


@libentry()
# @triton.autotune(configs=runtime.get_tuned_config("count_nonzero"), key=["numel"])
@triton.heuristics(
    {
        "BLOCK_SIZE": heur_block_size,
    }
)
@triton.jit
def count_nonzero_kernel(x_ptr, out_ptr, N, numel, BLOCK_SIZE: tl.constexpr):
    pid_x = tle.program_id(0)

    nonzero_count = tl.full((), value=0, dtype=out_ptr.dtype.element_ty)
    for start_n in range(0, N, BLOCK_SIZE):
        cols_offsets = start_n + tl.arange(0, BLOCK_SIZE)
        offset = pid_x * N + cols_offsets
        mask = offset < numel and cols_offsets < N
        x = tl.load(x_ptr + offset, mask=mask, other=0)
        is_nonzero = (x != 0).to(tl.int64)
        nonzero_count += tl.sum(is_nonzero)

    tl.store(out_ptr + pid_x, nonzero_count)


"""***************************** TROTITON XPU KERNEL *****************************"""


@libentry()
@triton.jit
def count_nonzero_kernel_xpu(
    x_ptr, out_ptr, M, N, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr
):
    pid_x = tl.program_id(0)
    row = pid_x * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    row_mask = row < M

    _nonzero_count = tl.zeros([BLOCK_M, BLOCK_N], dtype=out_ptr.dtype.element_ty)
    for start_n in range(0, N, BLOCK_N):
        cols = start_n + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask
        x = tl.load(x_ptr + row * N + cols, mask=mask, other=0)
        is_nonzero = (x != 0).to(tl.int64)
        _nonzero_count += is_nonzero

    nonzero_count = tl.sum(_nonzero_count, axis=1)[:, None]
    tl.store(out_ptr + row, nonzero_count, row_mask)


"""***************************** TROTITON XPU KERNEL *****************************"""


@libentry()
# @triton.autotune(configs=runtime.get_tuned_config("count_nonzero"), key=["numel"])
@triton.heuristics(
    {
        "BLOCK_SIZE": heur_block_size,
    }
)
@triton.jit
def count_nonzero_combin_kernel_1(x_ptr, out_ptr, N, numel, BLOCK_SIZE: tl.constexpr):
    pid_x = tle.program_id(0)
    nonzero_count = tl.full((), value=0, dtype=out_ptr.dtype.element_ty)
    for start_n in range(0, N, BLOCK_SIZE):
        cols_offsets = start_n + tl.arange(0, BLOCK_SIZE)
        offset = pid_x * N + cols_offsets
        mask = offset < numel and cols_offsets < N
        x = tl.load(x_ptr + offset, mask=mask, other=0)
        nonzero_count += tl.sum(x)
    tl.store(out_ptr + pid_x, nonzero_count)


@libentry()
@triton.jit
def count_nonzero_combin_kernel(
    x_ptr,
    combin_ptr,
    N: tl.constexpr,
    combin_N: tl.constexpr,
    numel: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    pid_x = tle.program_id(0)
    pid_y = tle.program_id(1)
    cols_offsets = pid_y * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    offset = pid_x * N + cols_offsets
    mask = offset < numel and cols_offsets < N
    x = tl.load(x_ptr + offset, mask=mask, other=0)
    is_nonzero = (x != 0).to(tl.int64)
    nonzero_count = tl.sum(is_nonzero)
    tl.store(combin_ptr + pid_x * combin_N + pid_y, nonzero_count)


def count_nonzero(x, dim=None):
    logger.debug("GEMS COUNT NONZERO")

    CORE_NUM = 64
    SIZE_PER_CORE = 512
    SIZE_PER_CLUSTER = CORE_NUM * SIZE_PER_CORE

    elem_bytes = x.element_size()
    if dim is not None:
        assert dim >= -x.ndim and dim < x.ndim, "Invalid dim"
        shape = x.shape
        numel = x.numel()
        # premute
        os.environ["TRITONXPU_IS_SCATTER_SLICE"] = "1"
        x = dim_compress(x, dim)
        x = x.contiguous().flatten()
        del os.environ["TRITONXPU_IS_SCATTER_SLICE"]
        # 2D count_nonzero
        out_shape = list(shape)
        del out_shape[dim]
        os.environ["TRITONXPU_ELEMBYTES"] = "8"
        out = torch.zeros(out_shape, dtype=torch.int64, device=x.device)
        del os.environ["TRITONXPU_ELEMBYTES"]
        N = shape[dim]
        M = triton.cdiv(numel, shape[dim])
        BLOCK_M = CORE_NUM
        BLOCK_N = SIZE_PER_CORE
        grid = lambda meta: (triton.cdiv(M, BLOCK_M),)
        os.environ["TRITONXPU_ELEMBYTES"] = "8"
        count_nonzero_kernel_xpu[grid](
            x,
            out,
            M,
            N,
            BLOCK_M=BLOCK_M,
            BLOCK_N=BLOCK_N,
            groups_per_cluster=CORE_NUM,
            buffer_size_limit=SIZE_PER_CORE * 8,
            is_use_mask_zero=True,
        )
        del os.environ["TRITONXPU_ELEMBYTES"]
        return out
    else:
        # 1D count_nonzero
        x = x.contiguous().flatten()
        numel = x.numel()
        gridX = triton.cdiv(numel, SIZE_PER_CLUSTER)
        os.environ["TRITONXPU_ELEMBYTES"] = "8"
        out_mid = torch.zeros(gridX, dtype=torch.int64, device=x.device)
        del os.environ["TRITONXPU_ELEMBYTES"]
        count_nonzero_kernel_1_part0_xpu[(gridX,)](
            x,
            out_mid,
            numel,
            BLOCK_SIZE_0=SIZE_PER_CLUSTER,
            buffer_size_limit=SIZE_PER_CORE * elem_bytes,
            is_use_mask_zero=True,
        )
        BLOCK_SIZE_1 = triton.next_power_of_2(gridX)
        os.environ["TRITONXPU_ELEMBYTES"] = "8"
        out = torch.zeros(1, dtype=torch.int64, device=x.device)
        count_nonzero_kernel_1_part1_xpu[(1,)](
            out_mid,
            out,
            gridX,
            BLOCK_SIZE_1=BLOCK_SIZE_1,
            buffer_size_limit=SIZE_PER_CORE * 8,
            is_use_mask_zero=True,
        )
        del os.environ["TRITONXPU_ELEMBYTES"]

        return out[0]



"""
neurobrix run — Execute inference using NBX Engine.

DATA-DRIVEN DESIGN:
- CLI does NO business logic
- All args are mapped to global.* variables
- Family is read from manifest.json (set at trace/import time)
- The model (via variables.json) decides how to use them

ZERO HARDCODE: Defaults cascade from CLI > runtime/defaults.json > family config
"""

import sys
import json
import time
from pathlib import Path

from neurobrix import __version__
from neurobrix.cli.utils import find_model


def _write_video_h264(output_path: str, frames_uint8, fps: float):
    """Write video as H.264/mp4 using ffmpeg (QuickTime/browser compatible)."""
    import subprocess
    import imageio_ffmpeg
    T, H, W, C = frames_uint8.shape
    ffmpeg_exe = imageio_ffmpeg.get_ffmpeg_exe()
    cmd = [
        ffmpeg_exe, "-y",
        "-f", "rawvideo",
        "-vcodec", "rawvideo",
        "-s", f"{W}x{H}",
        "-pix_fmt", "rgb24",
        "-r", str(fps),
        "-i", "-",
        "-vcodec", "libx264",
        "-pix_fmt", "yuv420p",
        "-crf", "18",
        "-preset", "medium",
        output_path,
    ]
    proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    proc.communicate(input=frames_uint8.tobytes())


def _try_warm_path(args) -> bool:
    """Attempt warm-path execution via running daemon. Returns True if handled."""
    from neurobrix.serving.client import DaemonClient

    if not DaemonClient.is_running():
        return False

    try:
        client = DaemonClient()
        client.connect()
        status = client.status()
    except Exception:
        return False

    # Only use warm path if daemon has the same model loaded
    if status.get("model") != args.model:
        client.close()
        return False

    print(f"[Run] Using warm daemon (PID {DaemonClient.get_pid()})")

    kwargs = {}
    if args.steps is not None:
        kwargs["steps"] = args.steps
    if args.height is not None:
        kwargs["height"] = args.height
    if args.width is not None:
        kwargs["width"] = args.width
    if args.cfg is not None:
        kwargs["cfg"] = args.cfg
    if args.temperature is not None:
        kwargs["temperature"] = args.temperature
    if args.repetition_penalty is not None:
        kwargs["repetition_penalty"] = args.repetition_penalty
    if args.chat_mode is not None:
        kwargs["chat_mode"] = args.chat_mode
    if args.seed is not None:
        kwargs["seed"] = args.seed

    # Audio models: pass audio_path to daemon
    if getattr(args, 'audio', None):
        kwargs["audio_path"] = args.audio

    # For non-LLM non-audio: pass output_path so daemon saves the file
    family = status.get("family")
    if family not in ("llm", "audio"):
        from neurobrix.core.config import get_output_processing
        _output_cfg = get_output_processing(family)
        _default_ext = _output_cfg.get("output_format", "png")
        output_path = args.output or f"output_{args.model}.{_default_ext}"
        # Resolve to absolute path so daemon saves in client's cwd, not daemon's
        kwargs["output_path"] = str(Path(output_path).resolve())

    try:
        result = client.generate(prompt=args.prompt or "", **kwargs)
        client.close()
    except RuntimeError as e:
        print(f"[Run] Daemon error: {e}")
        client.close()
        return False

    # Display result
    timing = result.get("timing", {})
    total_s = timing.get("total_s", 0)

    if family == "llm":
        text = result.get("text", "")
        tokens = result.get("tokens", 0)
        print(f"\n[Output] Generated {tokens} tokens in {total_s}s")
        if text:
            print(f"\n{text}")
    elif family == "audio":
        # Audio STT: transcription text
        transcription = result.get("transcription")
        if transcription:
            print(f"\n[Transcription] {transcription}")
        # Audio TTS: waveform saved
        saved_path = result.get("output_path")
        if saved_path:
            print(f"\n{'='*70}")
            print(f"SAVED: {saved_path}")
            print(f"{'='*70}")
    else:
        saved_path = result.get("output_path")
        if saved_path:
            print(f"\n{'='*70}")
            print(f"SAVED: {saved_path}")
            print(f"{'='*70}")
        else:
            print(f"\n[Output] Generation complete in {total_s}s")

    return True


def cmd_run(args):
    """Generate output using NeuroBrix Runtime."""
    import torch
    from neurobrix.nbx import NBXContainer
    from neurobrix.core.prism import PrismSolver, load_profile, InputConfig
    from neurobrix.core.prism.autodetect import get_or_create_default_profile
    from neurobrix.core.runtime.loader import NBXRuntimeLoader
    from neurobrix.core.runtime.executor import RuntimeExecutor
    from neurobrix.core.config import get_output_processing

    # Auto-detect model from running daemon if --model omitted
    if args.model is None:
        from neurobrix.serving.client import DaemonClient
        if DaemonClient.is_running():
            try:
                client = DaemonClient()
                client.connect()
                status = client.status()
                args.model = status.get("model")
                client.close()
            except Exception:
                pass
        if args.model is None:
            print("ERROR: --model is required when no daemon is running.")
            sys.exit(1)

    # Warm path: if daemon is running with same model, use it
    if _try_warm_path(args):
        sys.exit(0)

    # GUARD: If daemon is running but warm path failed, refuse cold path.
    # Cold run + daemon = double GPU allocation = OOM. One task at a time.
    from neurobrix.serving.client import DaemonClient
    if DaemonClient.is_running():
        daemon_pid = DaemonClient.get_pid()
        print(f"ERROR: A serving daemon is already running (PID {daemon_pid}).")
        print(f"NeuroBrix runs one task at a time — the daemon is using the GPUs.")
        print(f"Either:")
        print(f"  1. Use the daemon:  neurobrix run --model {args.model} --prompt '...'")
        print(f"  2. Stop it first:   neurobrix stop")
        sys.exit(1)

    # DATA-DRIVEN: Find model by scanning all family directories
    try:
        nbx_path = find_model(args.model)
    except FileNotFoundError as e:
        print(f"ERROR: {e}")
        sys.exit(1)

    print("=" * 70)
    print(f"NeuroBrix Run v{__version__}")
    print("=" * 70)
    print(f"Model: {args.model}")
    print(f"Hardware: {args.hardware or 'auto-detect'}")
    if args.prompt:
        prompt_display = f"{args.prompt[:50]}..." if len(args.prompt) > 50 else args.prompt
        print(f"Prompt: {prompt_display}")
    if getattr(args, 'audio', None):
        print(f"Audio: {args.audio}")
    print("=" * 70)

    # 1. Load NBX Container
    print("\n[1/4] Loading NBX container...")
    container = NBXContainer.load(str(nbx_path))

    # DATA-DRIVEN: Validate input modality before proceeding
    manifest = container.get_manifest() or {}
    family = manifest.get("family")
    input_modality = manifest.get("input_modality", "text")

    # Image-to-image models (super-resolution, upscaling)
    if input_modality == "image":
        model_name = manifest.get("model_name", args.model)
        print(f"\nERROR: '{model_name}' is an image-to-image model requiring --input-image.")
        print(f"Image-to-image inference is not yet supported.")
        sys.exit(1)

    # Audio family input validation
    if family == "audio":
        audio_arg = getattr(args, 'audio', None)
        if not audio_arg and not args.prompt:
            print("ERROR: Audio model requires --audio <file> (STT) or --prompt <text> (TTS)")
            sys.exit(1)
    elif not args.prompt:
        print("ERROR: --prompt is required for this model family.")
        sys.exit(1)

    neural_components = container.get_neural_components()
    print(f"   Components: {[c.name for c in neural_components]}")

    # 2. Prism Allocation
    print("\n[2/4] Solving hardware allocation...")
    if args.hardware:
        hw_profile = load_profile(args.hardware)
    else:
        hardware_id = get_or_create_default_profile()
        hw_profile = load_profile(hardware_id)
    print(f"   Profile: {hw_profile.id} ({hw_profile.total_vram_gb:.1f} GB)")

    # Build InputConfig for activation profiling
    cache_path = container._cache_path
    assert cache_path is not None, "Container cache path must be set"
    defaults_path = cache_path / "runtime" / "defaults.json"
    if defaults_path.exists():
        with open(defaults_path) as f:
            cached_defaults = json.load(f)
    else:
        cached_defaults = {}

    height = getattr(args, 'height', None) or cached_defaults.get("height", 1024)
    width = getattr(args, 'width', None) or cached_defaults.get("width", 1024)
    vae_scale = cached_defaults.get("vae_scale_factor", 8)

    input_config = InputConfig(
        batch_size=2,  # CFG effectively doubles batch
        height=height,
        width=width,
        dtype="float16",
        vae_scale=vae_scale,
    )

    solver = PrismSolver()
    execution_plan = solver.solve_smart(container, hw_profile, input_config)

    # Apply CPU optimizations from hardware profile
    if hw_profile.cpu:
        from neurobrix.core.prism.cpu_config import apply_cpu_config
        apply_cpu_config(
            cpu=hw_profile.cpu,
            strategy=execution_plan.strategy,
            device_count=hw_profile.device_count,
            preferred_dtype=hw_profile.preferred_dtype,
        )

    print(f"   Strategy: {execution_plan.strategy}")
    for comp_name, alloc in execution_plan.components.items():
        print(f"   {comp_name} → {alloc.device}")

    # 3. Load RuntimePackage
    print("\n[3/4] Loading runtime...")
    loader = NBXRuntimeLoader()
    pkg = loader.load(str(nbx_path))

    # 4. Build Universal Inputs Dictionary
    print("\n[4/4] Preparing inputs...")

    inputs = {}
    if args.prompt:
        inputs["global.prompt"] = args.prompt
    if getattr(args, 'audio', None):
        inputs["global.audio_path"] = args.audio

    if args.steps is not None:
        inputs["global.num_inference_steps"] = args.steps
    if args.height is not None:
        inputs["global.height"] = args.height
    if args.width is not None:
        inputs["global.width"] = args.width
    if args.cfg is not None:
        inputs["global.guidance_scale"] = args.cfg
    if args.temperature is not None:
        inputs["global.temperature"] = args.temperature
    if args.repetition_penalty is not None:
        inputs["global.repetition_penalty"] = args.repetition_penalty
    if args.chat_mode is not None:
        inputs["global.chat_mode"] = args.chat_mode

    print(f"   CLI inputs: {list(inputs.keys())}")

    if args.seed is not None:
        inputs["global.seed"] = args.seed
        torch.manual_seed(args.seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(args.seed)
        print(f"   global.seed = {args.seed}")

    # Universal --set injection
    if args.set:
        for item in args.set:
            if "=" not in item:
                print(f"   [WARNING] Invalid --set format: {item} (expected key=value)")
                continue
            key, value = item.split("=", 1)
            if value.lower() in ('true', 'false'):
                value = value.lower() == 'true'
            elif value.isdigit():
                value = int(value)
            else:
                try:
                    value = float(value)
                except ValueError:
                    pass
            inputs[key] = value
            print(f"   {key} = {value}")

    print(f"   Total inputs: {len(inputs)}")

    # 5. Determine Execution Engine Mode
    if args.sequential:
        execution_mode = "native"
    elif args.triton:
        execution_mode = "triton"
    else:
        execution_mode = "compiled"

    # 6. Execute
    print("\n[Execute] Running pipeline...")
    print(f"   Engine: {execution_mode.upper()}")
    executor = RuntimeExecutor(pkg, execution_plan, mode=execution_mode)

    try:
        if torch.cuda.is_available():
            torch.cuda.synchronize()
        t_exec_start = time.time()
        outputs = executor.execute(inputs)
        if torch.cuda.is_available():
            torch.cuda.synchronize()
        t_exec_total = time.time() - t_exec_start
    except Exception as e:
        print(f"\n[ERROR] Pipeline failed: {e}")
        import traceback
        traceback.print_exc()
        sys.exit(1)

    # 6a. Print timing summary
    print(f"\n[Timing] Total execution: {t_exec_total:.2f}s")

    # 7. Find and Save Output - FAMILY-DRIVEN
    family = pkg.manifest.get("family")
    if family is None:
        raise RuntimeError(
            f"ZERO FALLBACK: 'family' missing in manifest for model '{args.model}'.\n"
            f"Model data incomplete. Re-import: neurobrix remove {args.model} && neurobrix import <org>/<model>"
        )

    # =========================================================================
    # LLM Family: Text/Token Output
    # =========================================================================
    if family == "llm":
        output_tokens = outputs.get("output_tokens")
        if output_tokens is None:
            output_tokens = outputs.get("global.output_tokens")
        if output_tokens is None:
            print(f"\n[WARNING] No output_tokens found")
            print(f"Available: {list(outputs.keys())}")
            sys.exit(1)

        print(f"\n[Output] Generated {output_tokens.shape[-1]} tokens")

        generated_text = None
        if "tokenizer" in executor.modules:
            tokenizer = executor.modules["tokenizer"]
            if hasattr(tokenizer, 'decode'):
                token_ids = output_tokens.flatten().tolist()
                generated_text = tokenizer.decode(token_ids, skip_special_tokens=True)

        if args.output:
            output_path = args.output
            if output_path.endswith(".json"):
                import json as json_mod
                result = {
                    "model": args.model,
                    "choices": [{
                        "message": {
                            "role": "assistant",
                            "content": generated_text or str(output_tokens.flatten().tolist()),
                        },
                        "finish_reason": "stop",
                    }],
                }
                with open(output_path, 'w') as f:
                    json_mod.dump(result, f, indent=2, ensure_ascii=False)
            else:
                with open(output_path, 'w') as f:
                    f.write(generated_text if generated_text else str(output_tokens.flatten().tolist()))
            print(f"\n[Success] Output saved to: {output_path}")
        else:
            if generated_text:
                print(f"\n{generated_text}")
            else:
                print(f"\n{output_tokens.flatten().tolist()}")

        sys.exit(0)

    # =========================================================================
    # Audio Family: Transcription (STT) or Waveform (TTS)
    # =========================================================================
    if family == "audio":
        # STT: text transcription
        transcription = outputs.get("global.transcription")
        if transcription:
            print(f"\n[Transcription]")
            print(f"\n{transcription}")
            if args.output:
                with open(args.output, 'w') as f:
                    f.write(transcription)
                print(f"\nSaved to: {args.output}")
            sys.exit(0)

        # TTS: waveform tensor
        waveform = outputs.get("global.output_audio")
        if waveform is not None:
            output_path = args.output or f"output_{args.model}.wav"
            from neurobrix.core.module.audio.output_processor import AudioOutputProcessor
            from neurobrix.core.config import get_output_processing
            audio_cfg = get_output_processing("audio")
            flow_sr = pkg.topology.get("flow", {}).get("audio", {}).get("sample_rate")
            sample_rate = flow_sr or pkg.defaults.get("sample_rate", audio_cfg.get("sample_rate", 24000))
            AudioOutputProcessor.save_waveform(waveform, output_path, sample_rate)
            print(f"\n{'='*70}")
            print(f"SAVED: {output_path}")
            print(f"{'='*70}")
            sys.exit(0)

        print(f"\n[WARNING] Audio model produced no transcription or waveform output")
        print(f"Available: {list(outputs.keys())}")
        sys.exit(1)

    # =========================================================================
    # Image/Video Family: Tensor Output
    # =========================================================================
    final_output = executor.get_final_output(outputs)
    if final_output is None:
        final_comp = executor.get_final_component()
        print(f"\n[WARNING] No tensor output from '{final_comp}'")
        print(f"Available: {[k for k in outputs.keys() if isinstance(outputs.get(k), torch.Tensor)]}")
        sys.exit(1)

    print(f"\n[Output] Final tensor: {final_output.shape}")
    print(f"[Output] Raw VAE output: min={final_output.min():.4f}, max={final_output.max():.4f}, mean={final_output.mean():.4f}")

    output_range = pkg.defaults.get("output_range")
    if output_range is None:
        output_cfg = get_output_processing(family)
        output_range = output_cfg.get("output_range", [-1.0, 1.0])

    output_cfg = get_output_processing(family)
    batch_axis = output_cfg.get("batch_axis", 0)
    channel_axis = output_cfg.get("channel_axis", 1)
    valid_channels = output_cfg.get("valid_channels", [1, 3, 4])
    bit_depth = output_cfg.get("bit_depth", 8)
    layout = output_cfg.get("layout", "CHW")

    from neurobrix.core.module.output_processor import OutputProcessor
    processor = OutputProcessor.from_package(pkg)

    # Remove batch dimension
    tensor = torch.select(final_output, batch_axis, 0).cpu().float()
    tensor = processor.process(tensor, output_range)
    tensor = tensor.clamp(0, 1)

    import numpy as np

    # Default output path uses family's output_format
    default_ext = output_cfg.get("output_format", "png")
    output_path = args.output or f"output_{args.model}.{default_ext}"

    # ── VIDEO OUTPUT (4D: TCHW or CTHW) ──
    if family == "video" and tensor.dim() == 4:
        if not output_path.endswith(".mp4"):
            output_path = output_path.rsplit(".", 1)[0] + ".mp4"

        fps = output_cfg.get("fps", 24)
        fps = pkg.defaults.get("fps", fps)

        # Normalize to [T, H, W, C] for cv2
        if layout == "CTHW":
            # [C, T, H, W] → [T, H, W, C]
            frames = tensor.permute(1, 2, 3, 0).numpy()
        elif layout == "TCHW":
            # [T, C, H, W] → [T, H, W, C]
            frames = tensor.permute(0, 2, 3, 1).numpy()
        else:
            # Fallback: assume CTHW (diffusers standard)
            frames = tensor.permute(1, 2, 3, 0).numpy()

        frames_uint8 = (frames * 255).astype(np.uint8)
        T, H, W, C = frames_uint8.shape

        _write_video_h264(output_path, frames_uint8, fps)

        print(f"\n{'='*70}")
        print(f"SAVED: {output_path} ({T} frames, {W}x{H}, {fps} fps)")
        print(f"{'='*70}")
        return 0

    # ── IMAGE OUTPUT (3D: CHW) ──
    output_path = args.output or f"output_{args.model}.png"

    if layout == "CHW" and tensor.dim() == 3:
        actual_channel_axis = channel_axis - 1 if batch_axis < channel_axis else channel_axis
        if tensor.shape[actual_channel_axis] in valid_channels:
            tensor = tensor.permute(1, 2, 0)

    from PIL import Image

    if bit_depth == 8:
        img_np = (tensor.numpy() * 255).astype(np.uint8)
    elif bit_depth == 16:
        img_np = (tensor.numpy() * 65535).astype(np.uint16)
    else:
        img_np = (tensor.numpy() * 255).astype(np.uint8)

    if img_np.shape[-1] == 1:
        img_np = img_np.squeeze(-1)

    Image.fromarray(img_np).save(output_path)

    print(f"\n{'='*70}")
    print(f"SAVED: {output_path}")
    print(f"{'='*70}")

    return 0



"""PATH diagnostic helper — guides users when `neurobrix` CLI isn't on PATH."""

import os
import sys
import shutil
import sysconfig


def get_scripts_dir() -> str:
    """Return the directory where pip installs console_scripts."""
    # User installs (pip install --user)
    user_scripts = sysconfig.get_path("scripts", f"{os.name}_user")
    # Global/venv installs
    global_scripts = sysconfig.get_path("scripts")

    # Check which one actually has neurobrix
    for d in (user_scripts, global_scripts):
        if d and os.path.isfile(os.path.join(d, "neurobrix")):
            return d

    # Fallback: return whichever exists
    return user_scripts or global_scripts or ""


def check_cli_on_path() -> None:
    """Print a helpful message if `neurobrix` isn't on PATH."""
    if shutil.which("neurobrix"):
        return  # All good

    scripts_dir = get_scripts_dir()
    if not scripts_dir:
        return

    print("\033[33m")  # yellow
    print("  neurobrix is installed but not on your PATH.")
    print(f"  Script location: {scripts_dir}")
    print()
    _print_path_fix(scripts_dir)
    print()
    print("  Or run directly with: python -m neurobrix <command>")
    print("\033[0m")  # reset


def _print_path_fix(scripts_dir: str) -> None:
    """Print OS-specific PATH instructions."""
    if sys.platform == "darwin":
        shell = os.environ.get("SHELL", "")
        if "zsh" in shell:
            rc = "~/.zshrc"
        elif "bash" in shell:
            rc = "~/.bash_profile"
        else:
            rc = "~/.zshrc"  # macOS default
        print(f"  To fix, run:")
        print(f"    echo 'export PATH=\"{scripts_dir}:$PATH\"' >> {rc}")
        print(f"    source {rc}")

    elif sys.platform == "win32":
        print(f"  To fix, run (PowerShell as Admin):")
        print(f'    [Environment]::SetEnvironmentVariable("PATH", "{scripts_dir};" + '
              f'[Environment]::GetEnvironmentVariable("PATH", "User"), "User")')
        print()
        print(f"  Or via Settings: search 'Environment Variables' > Path > New > {scripts_dir}")

    else:  # Linux
        shell = os.environ.get("SHELL", "")
        if "zsh" in shell:
            rc = "~/.zshrc"
        elif "fish" in shell:
            rc = "~/.config/fish/config.fish"
            print(f"  To fix, run:")
            print(f"    fish_add_path {scripts_dir}")
            return
        else:
            rc = "~/.bashrc"
        print(f"  To fix, run:")
        print(f"    echo 'export PATH=\"{scripts_dir}:$PATH\"' >> {rc}")
        print(f"    source {rc}")


def print_path_diagnostics() -> None:
    """Full diagnostic output for `neurobrix doctor`."""
    print(f"Python:          {sys.executable}")
    print(f"Version:         {sys.version.split()[0]}")
    print(f"Platform:        {sys.platform}")
    print(f"Scripts dir:     {get_scripts_dir()}")
    print(f"neurobrix found: {shutil.which('neurobrix') or 'NOT ON PATH'}")
    print()

    if shutil.which("neurobrix"):
        print("neurobrix is on PATH. No action needed.")
    else:
        check_cli_on_path()



import logging
import math
from typing import Sequence

import torch
import triton
import triton.language as tl

from flag_gems.runtime import torch_device_fn
from flag_gems.utils import dim_compress, libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(
    f'flag_gems.runtime.backend._mthreads.ops.{__name__.split(".")[-1]}'
)


@triton.jit
def reduce_mul(a, b):
    return a * b


NAIVE_REDUCTION_CONFIGS = [
    triton.Config({"BLOCK_M": 8, "BLOCK_N": 64}, num_warps=2),
    triton.Config({"BLOCK_M": 16, "BLOCK_N": 128}, num_warps=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 128}, num_warps=4),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 128}, num_warps=4, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 256}, num_warps=4, num_stages=2),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 256}, num_warps=8, num_stages=2),
    triton.Config({"BLOCK_M": 128, "BLOCK_N": 256}, num_warps=8, num_stages=2),
    triton.Config({"BLOCK_M": 32, "BLOCK_N": 512}, num_warps=8, num_stages=2),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 512}, num_warps=8, num_stages=2),
    triton.Config({"BLOCK_M": 64, "BLOCK_N": 1024}, num_warps=8, num_stages=2),
]


def _prune_reduction_configs(configs, named_args, **meta):
    """Skip oversized tiles to avoid needless autotune on tiny shapes."""
    M = named_args["M"]
    N = named_args["N"]
    max_block_m = max(M, 8)
    min_block_m = 8
    n_cap = 1 << (N - 1).bit_length()
    n_cap = max(64, min(n_cap, 1024))
    filtered = [
        cfg
        for cfg in configs
        if min_block_m <= cfg.kwargs["BLOCK_M"] <= max_block_m
        and cfg.kwargs["BLOCK_N"] <= max(256, n_cap)
    ]
    return filtered or configs


def _flatten_dim(shape: Sequence[int], dim: int):
    dim = dim % len(shape)
    n = shape[dim]
    inner = math.prod(shape[dim + 1 :]) if dim + 1 < len(shape) else 1
    outer = math.prod(shape[:dim]) if dim > 0 else 1
    return dim, n, inner, outer


def _reshape_output(out: torch.Tensor, shape: list[int], dim: int, keepdim: bool):
    out_shape = shape.copy()
    out_shape[dim] = 1
    out_view = out.view(out_shape)
    if not keepdim:
        out_view = torch.squeeze(out_view, dim)
    return out_view


@libentry()
@triton.jit
def prod_kernel_mid(
    inp,
    mid,
    M,
    BLOCK_SIZE: tl.constexpr,
):
    dtype = inp.type.element_ty
    acc_dtype = tl.float32
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < M
    inp_val = tl.load(inp_ptrs, mask=mask, other=1.0).to(acc_dtype)
    mid_value = tl.reduce(inp_val, axis=0, combine_fn=reduce_mul).to(dtype)
    mid_ptr = mid + pid
    tl.store(mid_ptr, mid_value)


@libentry()
@triton.jit
def prod_kernel_result(mid, out, mid_size, BLOCK_MID: tl.constexpr):
    dtype = mid.type.element_ty
    acc_dtype = tl.float32
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid + offset
    mask = offset < mid_size
    mid_val = tl.load(mid_ptrs, mask=mask, other=1.0).to(acc_dtype)
    prod_val = tl.reduce(mid_val, axis=0, combine_fn=reduce_mul).to(dtype)
    tl.store(out, prod_val)


@triton.jit
def prod_kernel_dim_64(
    inp,
    out,
    M,
    INNER,
    STRIDE_OUTER,
    BLOCK_M: tl.constexpr,
):
    pid = tle.program_id(0)
    rows = pid * BLOCK_M + tl.arange(0, BLOCK_M)
    row_mask = rows < M
    base_ptr = inp + rows * STRIDE_OUTER
    cols = tl.arange(0, 64)
    vals = tl.load(base_ptr[:, None] + cols[None, :], cache_modifier=".cg")
    prod_vals = tl.reduce(vals.to(tl.float32), axis=1, combine_fn=reduce_mul)
    tl.store(out + rows, prod_vals.to(inp.type.element_ty), mask=row_mask)


@triton.jit
def prod_kernel_dim_contig(
    inp,
    out,
    M,
    INNER,
    STRIDE_OUTER,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    pid = tle.program_id(0)
    rows = pid * BLOCK_M + tl.arange(0, BLOCK_M)
    row_mask = rows < M
    base_ptr = inp + rows * STRIDE_OUTER
    cols = tl.arange(0, BLOCK_N)
    col_mask = cols[None, :] < STRIDE_OUTER
    mask = row_mask[:, None] & col_mask
    vals = tl.load(
        base_ptr[:, None] + cols[None, :],
        mask=mask,
        other=1.0,
        cache_modifier=".cg",
    )
    prod_vals = tl.reduce(vals.to(tl.float32), axis=1, combine_fn=reduce_mul)
    tl.store(out + rows, prod_vals.to(inp.type.element_ty), mask=row_mask)


@triton.jit
def prod_kernel_dim_dense(
    inp,
    out,
    M,
    N,
    INNER,
    STRIDE_OUTER,
    STRIDE_REDUCE,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    dtype = inp.type.element_ty
    acc_dtype = tl.float32
    pid = tle.program_id(0)
    rows = pid * BLOCK_M + tl.arange(0, BLOCK_M)
    outer_idx = rows // INNER
    inner_idx = rows % INNER
    base_ptr = inp + outer_idx * STRIDE_OUTER + inner_idx

    acc = tl.full((BLOCK_M,), value=1.0, dtype=acc_dtype)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)
        vals = tl.load(
            base_ptr[:, None] + cols[None, :] * STRIDE_REDUCE,
            cache_modifier=".cg",
        ).to(acc_dtype)
        chunk_prod = tl.reduce(vals, axis=1, combine_fn=reduce_mul)
        acc *= chunk_prod

    tl.store(out + rows, acc.to(dtype))


@triton.autotune(
    configs=NAIVE_REDUCTION_CONFIGS,
    key=["M", "N"],
    prune_configs_by={"early_config_prune": _prune_reduction_configs},
    warmup=2,
    rep=8,
)
@triton.jit
def prod_kernel_dim(
    inp,
    out,
    M,
    N,
    INNER,
    STRIDE_OUTER,
    STRIDE_REDUCE,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    dtype = inp.type.element_ty
    acc_dtype = tl.float32
    pid = tle.program_id(0)
    rows = pid * BLOCK_M + tl.arange(0, BLOCK_M)
    rows = rows.to(tl.int64)
    row_mask = rows < M

    outer_idx = rows // INNER
    inner_idx = rows % INNER
    base_ptr = inp + outer_idx * STRIDE_OUTER + inner_idx

    acc = tl.full((BLOCK_M,), value=1.0, dtype=acc_dtype)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)
        cols = cols.to(tl.int64)
        col_mask = cols < N
        mask = row_mask[:, None] & col_mask[None, :]
        vals = tl.load(
            base_ptr[:, None] + cols[None, :] * STRIDE_REDUCE,
            mask=mask,
            other=1.0,
            cache_modifier=".cg",
        ).to(acc_dtype)
        chunk_prod = tl.reduce(vals, axis=1, combine_fn=reduce_mul)
        acc *= chunk_prod

    out_ptrs = out + rows
    tl.store(out_ptrs, acc.to(dtype), mask=row_mask)


def prod(inp, *, dtype=None):
    logger.debug("GEMS_MTHREADS PROD")
    if dtype is None:
        dtype = inp.dtype
    if not inp.is_contiguous():
        inp = inp.contiguous()

    M = inp.numel()
    block_size = triton.next_power_of_2(math.ceil(math.sqrt(M)))
    block_size = min(block_size * 2, 4096, triton.next_power_of_2(M))
    mid_size = triton.cdiv(M, block_size)
    block_mid = triton.next_power_of_2(mid_size)

    mid = torch.empty((mid_size,), dtype=dtype, device=inp.device)
    out = torch.empty([], dtype=dtype, device=inp.device)

    with torch_device_fn.device(inp.device):
        prod_kernel_mid[(mid_size, 1, 1)](inp, mid, M, block_size)
        prod_kernel_result[(1, 1, 1)](mid, out, mid_size, block_mid)
    return out


def prod_dim(inp, dim=None, keepdim=False, *, dtype=None):
    logger.debug("GEMS_MTHREADS PROD DIM")
    assert dim is not None, "dim must be specified"
    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    dim = dim % inp.ndim

    if dtype is None:
        dtype = inp.dtype
    if not inp.is_contiguous():
        inp = dim_compress(inp, dim)
        dim = inp.ndim - 1

    shape = list(inp.shape)
    dim, n, inner, outer = _flatten_dim(shape, dim)
    m = outer * inner

    out_flat = torch.empty((m,), dtype=dtype, device=inp.device)

    stride = inp.stride()
    stride_reduce = stride[dim]
    stride_outer = stride_reduce * n

    if n == 64 and stride_reduce == 1 and stride_outer == n:
        grid_64 = (triton.cdiv(m, 8),)
        with torch_device_fn.device(inp.device):
            prod_kernel_dim_64[grid_64](
                inp, out_flat, m, inner, stride_outer, BLOCK_M=8, num_warps=2
            )
        return _reshape_output(out_flat, shape, dim, keepdim)

    key = (m, n, str(dtype), str(out_flat.dtype))
    config = prod_kernel_dim.cache.get(key, None)
    if m * n >= 64 * 1024 * 1024 and config is None:
        if dtype in (torch.float16, torch.bfloat16):
            config = triton.Config(
                {"BLOCK_M": 64, "BLOCK_N": 256}, num_warps=8, num_stages=2
            )
        else:
            config = triton.Config(
                {"BLOCK_M": 64, "BLOCK_N": 128}, num_warps=4, num_stages=1
            )
        prod_kernel_dim.cache[key] = config

    if config is not None:
        block_m_cfg = config.kwargs["BLOCK_M"]
        block_n_cfg = config.kwargs["BLOCK_N"]
        if m % block_m_cfg == 0 and n % block_n_cfg == 0:
            grid_dense = (m // block_m_cfg,)
            with torch_device_fn.device(inp.device):
                prod_kernel_dim_dense[grid_dense](
                    inp,
                    out_flat,
                    m,
                    n,
                    inner,
                    stride_outer,
                    stride_reduce,
                    BLOCK_M=block_m_cfg,
                    BLOCK_N=block_n_cfg,
                    num_warps=config.num_warps or 4,
                    num_stages=config.num_stages or 1,
                )
            return _reshape_output(out_flat, shape, dim, keepdim)

    if stride_reduce == 1 and stride_outer == n and n <= 1024:
        block_m = 128 if n >= 256 else 64
        block_n = min(512, max(64, 1 << (n - 1).bit_length()))
        grid_contig = (triton.cdiv(m, block_m),)
        with torch_device_fn.device(inp.device):
            prod_kernel_dim_contig[grid_contig](
                inp,
                out_flat,
                m,
                inner,
                stride_outer,
                BLOCK_M=block_m,
                BLOCK_N=block_n,
                num_warps=8 if n >= 256 else 4,
                num_stages=2,
            )
        return _reshape_output(out_flat, shape, dim, keepdim)

    if n <= 64:
        prod_kernel_dim.cache[key] = triton.Config(
            {"BLOCK_M": 8, "BLOCK_N": 64}, num_warps=2, num_stages=1
        )

    grid = lambda meta: (triton.cdiv(m, meta["BLOCK_M"]),)
    with torch_device_fn.device(inp.device):
        prod_kernel_dim[grid](
            inp,
            out_flat,
            m,
            n,
            max(inner, 1),
            stride_outer,
            stride_reduce,
        )

    return _reshape_output(out_flat, shape, dim, keepdim)



import logging
import math

import torch
import triton
import triton.language as tl

from ..utils import libentry
from ..utils import triton_lang_extension as tle

logger = logging.getLogger(__name__)


@libentry()
@triton.jit
def logspace_kernel(
    out_ptr,
    out_stride0,
    start,
    step_size,
    steps,
    log2_base: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    idx = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = idx < steps

    exponent = start + idx * step_size
    vals = tl.exp2(log2_base * exponent)

    tl.store(out_ptr + idx * out_stride0, vals, mask=mask)


def logspace(
    start,
    end,
    steps,
    base=10.0,
    *,
    dtype=None,
    layout=None,
    device=None,
    pin_memory=None,
) -> torch.Tensor:
    logger.debug("GEMS LOGSPACE")
    assert steps >= 0, "number of steps must be non-negative"
    out_dtype = dtype if dtype is not None else torch.get_default_dtype()
    out = torch.empty(
        steps,
        dtype=out_dtype,
        layout=layout,
        device=device,
        pin_memory=pin_memory,
    )
    if steps == 0:
        pass
    elif steps == 1:
        if isinstance(start, torch.Tensor):
            start = start.item()
        out = torch.fill(out, base**start)
    else:
        if isinstance(start, torch.Tensor):
            start = start.item()
        if isinstance(end, torch.Tensor):
            end = end.item()
        step_size = (float(end) - float(start)) / (steps - 1)
        BLOCK_SIZE = 256  # according to benchmark, 256 is the best block size
        grid = (triton.cdiv(steps, BLOCK_SIZE),)
        logspace_kernel[grid](
            out,
            out.stride(0),
            start,
            step_size,
            steps,
            log2_base=math.log2(float(base)),  # math.log2 require float input
            BLOCK_SIZE=BLOCK_SIZE,
        )

    return out



# Softmax - Pure Triton
# NeuroBrix - NVIDIA Common (All architectures)

import torch
import triton
import triton.language as tl

@triton.jit
def _softmax_kernel(
    output_ptr, input_ptr,
    input_row_stride, output_row_stride,
    n_cols,
    BLOCK_SIZE: tl.constexpr
):
    """Softmax along last axis - pure Triton."""
    row_idx = tl.program_id(0)
    row_start_ptr = input_ptr + row_idx * input_row_stride
    col_offsets = tl.arange(0, BLOCK_SIZE)
    input_ptrs = row_start_ptr + col_offsets

    mask = col_offsets < n_cols
    row = tl.load(input_ptrs, mask=mask, other=-float('inf'))

    row_minus_max = row - tl.max(row, axis=0)
    numerator = tl.exp(row_minus_max)
    denominator = tl.sum(numerator, axis=0)
    softmax_output = numerator / denominator

    output_row_start_ptr = output_ptr + row_idx * output_row_stride
    output_ptrs = output_row_start_ptr + col_offsets
    tl.store(output_ptrs, softmax_output, mask=mask)



"""
Kernels for dropout.
"""


import triton
import triton.language as tl

from .utils import element_wise_kernel_configs


@triton.jit
def apply_dropout(input, drop_p, seed, offset):
    """
    Randomly zeroes elements in the input.

    Args:
        input: Input. The input must be loaded and cannot be a pointer.
        drop_p: Probability of dropping an element.
        seed: Seed for generating the dropout mask.
        offset: Offset to generate the mask for.

    Returns:
        Input with elements randomly zeroed out.
    """
    random = tl.rand(seed, offset)
    return tl.where(random < drop_p, 0, input / (1 - drop_p))


@triton.jit
def apply_dropout_grad(output_grad, drop_p, seed, offset):
    """
    Calculates the input gradient of dropout.

    Args:
        output_grad: Output gradients. The output gradients must be
            loaded and cannot be a pointer.
        drop_p: Probability of dropping an element.
        seed: Seed for generating the dropout mask.
        offset: Offset to generate the mask for.

    Returns:
        Gradient of dropout.
    """
    random = tl.rand(seed, offset)
    return tl.where(random < drop_p, 0, output_grad / (1 - drop_p))


@triton.autotune(
    configs=element_wise_kernel_configs(),
    key=['size'],
)
@triton.jit
def dropout_forward_kernel(
    input_pointer, output_pointer, size,
    drop_p, seed,
    BLOCK_SIZE: tl.constexpr,
    ):
    """
    Randomly zeroes elements in the input.

    Args:
        input_pointer: Pointer to the input to perform dropout on.
            The input must be of shape [size].
        output_pointer: Pointer to a container the result is written to.
            The container must be of shape [size].
        size: Number of elements in the input.
        drop_p: Probability of dropping an element.
        seed: Seed for generating the dropout mask.
        BLOCK_SIZE: Block size.
    """
    # This program processes BLOCK_SIZE rows.
    pid = tl.program_id(axis=0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offset < size

    input = tl.load(input_pointer + offset, mask=mask)
    output = apply_dropout(input, drop_p, seed, offset)
    tl.store(output_pointer + offset, output, mask=mask)


@triton.autotune(
    configs=element_wise_kernel_configs(),
    key=['size'],
)
@triton.jit
def dropout_backward_kernel(
    output_grad_pointer, input_grad_pointer, size,
    drop_p, seed,
    BLOCK_SIZE: tl.constexpr,
    ):
    """
    Calculates the input gradient of dropout.

    Args:
        output_grad_pointer: Pointer to dropout's output gradients.
            The output gradients must be of shape [size].
        input_grad_pointer: Pointer to a container the input's gradients are written to.
            The container must be of shape [size].
        size: Number of elements in the input.
        drop_p: Probability of dropping an element used in dropout.
        seed: Seed for generating the dropout mask.
        BLOCK_SIZE: Block size.
    """
    # This program processes BLOCK_SIZE rows.
    pid = tl.program_id(axis=0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offset < size

    output_grad = tl.load(output_grad_pointer + offset, mask=mask)
    input_grad = apply_dropout_grad(output_grad, drop_p, seed, offset)
    tl.store(input_grad_pointer + offset, input_grad, mask=mask)



import logging

import triton
import triton.language as tl

from flag_gems.utils import pointwise_dynamic, tl_extra_shim

_isinf = tl_extra_shim.isinf

logger = logging.getLogger(__name__)


@pointwise_dynamic(promotion_methods=[(0, "ALWAYS_BOOL")])
@triton.jit
def isinf_func(x):
    return _isinf(x.to(tl.float32))


def isinf(A):
    logger.debug("GEMS ISINF")
    return isinf_func(A)
